import numpy as np
import scipy as sp
import cuqi
import pytest
from cuqi.geometry import Continuous2D, Continuous1D


@pytest.mark.parametrize(
    "geomClass,grid,expected_grid,expected_par_shape,expected_fun_shape,expected_dim",
    [(cuqi.geometry.Continuous1D, (1), np.array([0]), (1,), (1,), 1),
     (cuqi.geometry.Continuous1D, (1,),
      np.array([0]), (1,), (1,), 1),
     (cuqi.geometry.Continuous1D, 1,
      np.array([0]), (1,), (1,), 1),
     (cuqi.geometry.Continuous1D, [1, 2, 3, 4], np.array(
         [1, 2, 3, 4]), (4,), (4,), 4),
     (cuqi.geometry.Continuous1D, 5, np.array(
         [0, 1, 2, 3, 4]), (5,), (5,), 5),
     (cuqi.geometry.Continuous2D, (1, 1),
      (np.array([0]), np.array([0])), (1,), (1, 1), 1),
     (cuqi.geometry.Continuous2D, ([1, 2, 3], 1), (np.array(
         [1, 2, 3]), np.array([0])), (3,), (3, 1), 3),
     (cuqi.geometry.Continuous2D, (2, 2), (np.array(
         [0, 1]), np.array([0, 1])), (4,), (2, 2), 4)
     ])
def test_Continuous_geometry(geomClass,grid,expected_grid,expected_par_shape, expected_fun_shape,expected_dim):
    geom = geomClass(grid=grid)
    assert(np.all(np.hstack(geom.grid) == np.hstack(expected_grid))
           and (geom.par_shape == expected_par_shape)
	   and (geom.par_dim == expected_dim)
	   and (geom.fun_shape == expected_fun_shape)
	   and (geom.fun_dim == expected_dim))

@pytest.mark.parametrize("geomClass",
                         [(cuqi.geometry.Continuous1D),
			  (cuqi.geometry.Continuous2D)])
def test_None_Continuous_geometry(geomClass):
    geom = geomClass()
    assert(    (geom.grid == None)
           and (geom.par_shape == None)
           and (geom.par_dim == None)
           and (geom.fun_shape == None)
	   and (geom.fun_dim == None))

@pytest.mark.parametrize("geomClass,grid,expected_grid,expected_shape,expected_dim",
                         [(cuqi.geometry.Continuous1D, (4,), np.array([0,1,2,3]),(4,),4),
			  (cuqi.geometry.Continuous2D, (2,3), (np.array([0,1]), np.array([0,1,2])), (2,3), 6)])
def test_update_Continuous_geometry(geomClass,grid,expected_grid,expected_shape,expected_dim):
    geom = geomClass()
    geom.grid = grid
    assert(np.all(np.hstack(geom.grid) == np.hstack(expected_grid))
           and (geom.fun_shape == expected_shape)
	   and (geom.fun_dim == expected_dim))

@pytest.mark.parametrize("variables,expected_variables,expected_shape,expected_dim",
                         [(3,['v0','v1','v2'],(3,),3),
			  (['a','b'],['a','b'],(2,),2),
			  (1,['v'],(1,),1),
			  ])
def test_Discrete_geometry(variables,expected_variables,expected_shape,expected_dim):
    geom = cuqi.geometry.Discrete(variables)
    assert(geom.variables == expected_variables
           and (geom.par_shape == expected_shape)
	   and (geom.par_dim == expected_dim))

@pytest.mark.parametrize("geom1,geom2,truth_value",
                         [(cuqi.geometry._DefaultGeometry1D(2),cuqi.geometry.Continuous1D(2), True),
			  (cuqi.geometry.Continuous1D(2),cuqi.geometry.Continuous2D((1,2)), False),
			  (cuqi.geometry._DefaultGeometry1D(np.array([0,1])),cuqi.geometry.Continuous1D(2), True),
			  (cuqi.geometry.Continuous1D(2),cuqi.geometry._DefaultGeometry1D(3), False),
			  (cuqi.geometry.Discrete(2),cuqi.geometry.Discrete(["v0","v1"]), True),
			  (cuqi.geometry.Discrete(2),cuqi.geometry.Continuous1D(2), False)])
def test_geometry_equivalence(geom1,geom2,truth_value):
    assert( (geom1==geom2) == truth_value)

# Make sure plotting does not fail at least
@pytest.mark.parametrize("is_par,plot_par",	[(True,False),(True,True),(False,False)])
@pytest.mark.parametrize("geom,val",[
						(cuqi.geometry.Discrete(1),np.pi,),
						(cuqi.geometry.Discrete(1),[np.pi]),
						(cuqi.geometry.Discrete(3),[1,2,3]),
						(cuqi.geometry.Continuous1D(1), np.pi,),
						(cuqi.geometry.Continuous1D(1),[np.pi]),
						(cuqi.geometry.Continuous1D(3),[1,2,3]),
						])
def test_geometry_plot(geom,val,is_par,plot_par):
	geom.plot(val,is_par=is_par,plot_par=plot_par)

@pytest.mark.parametrize("is_par,plot_par",	[(True,False),(True,True),(False,False)])
@pytest.mark.parametrize("geom,lo_val,hi_val",[
						(cuqi.geometry.Discrete(1),np.pi,2*np.pi),
						(cuqi.geometry.Discrete(1),[np.pi],[2*np.pi]),
						(cuqi.geometry.Discrete(3),[1,2,3],[4,5,6]),
						(cuqi.geometry.Continuous1D(1),np.pi,2*np.pi),
						(cuqi.geometry.Continuous1D(1),[np.pi],[2*np.pi]),
						(cuqi.geometry.Continuous1D(3),[1,2,3],[4,5,6]),
						])
def test_geometry_plot(geom,lo_val,hi_val,is_par,plot_par):
	geom.plot_envelope(lo_val,hi_val,is_par=is_par,plot_par=plot_par)

def test_geometry_variables_generator_default():
	g1 = cuqi.geometry._DefaultGeometry1D(5)
	g2 = cuqi.geometry._DefaultGeometry1D(5)
	g1.variables #Extract variables (they are generated in g1, but not in g2)
	assert g1==g2 #g2 has no _variables yet, but during check its generated.

def test_geometry_variables_generator_Geometry():
	g1 = cuqi.geometry.Continuous1D(5)
	g2 = cuqi.geometry.Continuous1D(5)
	g1.variables #Extract variables (they are generated in g1, but not in g2)
	assert g1==g2 #g2 has no _variables yet, but during check its generated.


@pytest.mark.parametrize("geom, map, imap",
                         [(cuqi.geometry.Discrete(4), lambda x:x**2, lambda x:np.sqrt(x)),
			  (cuqi.geometry.Continuous1D(15),lambda x:x+12, lambda x:x-12),
			  (cuqi.geometry.Continuous2D((4,5)), lambda x:x**2+1, lambda x:np.sqrt(x-1))
			  ])
def test_mapped_geometry(geom, map, imap):
    np.random.seed(0)

    mapped_geom = cuqi.geometry.MappedGeometry(geom, map, imap)
    val = np.random.rand(mapped_geom.par_dim)
    mapped_val = mapped_geom.par2fun(val)
    imapped_mapped_val = mapped_geom.fun2par(mapped_val)
    assert(np.allclose(val, imapped_mapped_val) and geom.par_shape == mapped_geom.par_shape)
    
@pytest.mark.parametrize("g1, g2, truth_value",[
						(Continuous2D((128,128)), Continuous2D((128,128)), True),
						(Continuous2D((1,2)), Continuous2D((1,2)), True),
						(Continuous2D((3,2)), Continuous2D((3,2)), True),
						(Continuous2D(), Continuous2D(), True),
						(Continuous1D(5), Continuous1D(5), True),
						(Continuous1D(), Continuous1D(), True),
						(Continuous2D((128,128)), Continuous2D((128,127)), False),
						(Continuous2D((1,2)), Continuous2D((1,3)), False),
						(Continuous2D((3,2)), Continuous2D((3,3)), False),
						(Continuous2D(), Continuous2D((1,2)), False),
						(Continuous2D((1,2)), Continuous2D(), False),
						(Continuous1D(5), Continuous1D(6), False),
						(Continuous1D(), Continuous1D(1), False),
						(Continuous1D(1), Continuous1D(), False),
						])
def test_geometry_equality(g1, g2, truth_value):
	"""Ensure geometry arrays compare correctly"""
	assert (g1==g2) == truth_value

@pytest.mark.parametrize("n_steps", [1, 2, 6, 7, 9, 10, 20, 21])
def test_StepExpansion_geometry(n_steps, copy_reference):
    """Check StepExpansion geometry correctness"""
    grid = np.linspace(0,1,20)
    if n_steps > np.size(grid):
	#If n_steps is greater than the number of grid points, StepExpansion will fail
        with pytest.raises(ValueError):
            cuqi.geometry.StepExpansion(grid,n_steps)
    else:
	#Otherwise, assert that the StepExpansion is correct
        geom = cuqi.geometry.StepExpansion(grid,n_steps)
        par = np.random.randn(n_steps)
        geom.plot(par,linestyle = '', marker='.')

        fun = geom.par2fun(par)
        par2 = geom.fun2par(fun)

        assert np.allclose(par, par2) \
           and geom.par_dim == n_steps
        
        # Assert fun and par2 matches the values in the data file
        ref_file = copy_reference(
            f"data/geometry/test_StepExpansion_{n_steps}.npz")
        ref = np.load(ref_file)
        assert np.allclose(fun, ref["fun"])
        assert np.allclose(par2, ref["par2"])

@pytest.mark.parametrize("projection, func",[('MiN', np.min),
      ('mAX', np.max),('mean', np.mean)])
def test_stepExpansion_fun2par(projection, func):
    """Check StepExpansion fun2par correctness when different projection methods are used"""

    # Set up geometry and grid
    np.random.seed(0)
    grid = np.linspace(0,10, 100, endpoint=True)
    n_steps = 3
    SE_geom = cuqi.geometry.StepExpansion(grid, n_steps=n_steps, fun2par_projection=projection)
    
    # Create cuqi array of function values
    qa_f = cuqi.array.CUQIarray(np.random.rand(len(grid)), is_par=False, geometry = SE_geom)
    
    # Compute projection manually (function value to parameters)
    p = np.empty(n_steps)
    p[0]= func(qa_f[np.where(grid <=grid[-1]/n_steps)])
    p[1]= func(qa_f[np.where((grid[-1]/n_steps < grid)&( grid <=2*grid[-1]/n_steps))])
    p[2]= func(qa_f[np.where(2*grid[-1]/n_steps < grid)])

    # Compare projection with fun2par results
    assert np.allclose(p, qa_f.parameters)

@pytest.mark.parametrize("num_modes",[1, 10, 20, 25])
def test_KL_expansion(num_modes, copy_reference):
    """Check KL expansion geometry correctness"""

    # File name for reference data
    ref_fname = f"data/geometry/KL_expansion_{num_modes}.npz"

    # Set up KL expansion geometry
    N = 20
    grid = np.linspace(0, 1, N)
    decay_rate = 2.5
    normalizer = 12
    geom = cuqi.geometry.KLExpansion(grid,
                                     decay_rate=decay_rate,
                                     normalizer=normalizer,
                                     num_modes=num_modes)
    if num_modes > len(grid):
        num_modes = len(grid)

    # Apply par2fun and check results
    p = np.random.randn(N)
    f_geom = geom.par2fun(p[:num_modes])

    p[num_modes:] = 0
    f_expected = _inverse_sin_discrete_transform_KL(
        p, N, decay_rate, normalizer)

    assert np.allclose(f_geom, f_expected)
    assert len(geom.coefs) == geom.par_dim
    assert geom.par_dim == geom.num_modes
    assert len(geom.grid) == geom.fun_dim

    # Verify the KL expansion results against the reference data
    ref_file = copy_reference(ref_fname)
    ref = np.load(ref_file)
    assert np.allclose(f_geom, ref["f_geom"])

def test_KLExpansion_set_grid():
    """Check updating grid in KL expansion geometry"""
    dim = 100
    grid = np.linspace(0, 1, dim)
    geom = cuqi.geometry.KLExpansion(grid, num_modes=200, decay_rate=1.5)

    # If num_modes > len(grid), num_modes is set to len(grid)
    assert geom.num_modes == 100 and len(geom.coefs) == 100

    # Update grid (num_modes > len(grid))
    geom.grid = np.linspace(0, 1, 120)
    assert geom.num_modes == 120 and len(geom.coefs) == 120

    # Update grid (num_modes < len(grid))
    geom.grid = np.linspace(0, 1, 300)
    assert geom.num_modes == 200 and len(geom.coefs) == 200


def test_KLExpansion_input():
    """Check KL expansion geometry par2fun input"""
    grid = np.linspace(0, 1, 100)
    geom = cuqi.geometry.KLExpansion(grid, num_modes=30, decay_rate=1.5)

    input = np.random.randn(150)

    # Correct input
    geom.par2fun(input[:30])

    # Input larger than num_modes
    with pytest.raises(ValueError):
        geom.par2fun(input)

    # Input smaller than num_modes
    with pytest.raises(ValueError):
        geom.par2fun(input[:5])


def test_KLExpansion_None_grid():
    """Check KL expansion geometry when grid is None at the initialization or
    set to None later"""

    # Initialize KLExpansion geometry with None grid
    geom = cuqi.geometry.KLExpansion(None, num_modes=100, decay_rate=1.5)

    # Check geometry properties
    assert geom.num_modes == 0
    assert geom.par_dim == 0
    assert geom.fun_dim is None
    assert geom.coefs is None

    # Set grid
    geom.grid = np.linspace(0, 1, 110)

    # Check geometry properties
    assert geom.num_modes == 100
    assert geom.par_dim == 100
    assert geom.fun_dim == 110
    assert geom.coefs is not None

    # Initialize KLExpansion geometry with grid of length 110
    geom = cuqi.geometry.KLExpansion(np.linspace(
        0, 1, 110), num_modes=100, decay_rate=1.5)

    # Check geometry properties
    assert geom.num_modes == 100
    assert geom.par_dim == 100
    assert geom.fun_dim == 110
    assert geom.coefs is not None

    # Set grid to None
    geom.grid = None

    # Check geometry properties
    assert geom.num_modes == 0
    assert geom.par_dim == 0
    assert geom.fun_dim is None
    assert geom.coefs is None

    # Initialize KLExpansion geometry with no arguments
    geom = cuqi.geometry.KLExpansion()

    # Check geometry properties
    assert geom.num_modes == 0
    assert geom.par_dim == 0
    assert geom.fun_dim is None
    assert geom.coefs is None

def _inverse_sin_discrete_transform_KL(p, N, decay_rate, normalizer):
    """Inverse of sin discrete transform, the code
    computes the KL expansion coefficients that the KLExpansion geometry computes given p and then computes the same transformation done by `idst` from
    `scipy.fftpack`"""
    p_f2 = np.zeros(N)
    modes = []
    K = np.arange(N)
    for i in range(0,N-1):

        coeff = 1/np.float_power( i+1,decay_rate )
        modes.append( p[i]*coeff/normalizer)
        p_f2 +=  modes[-1]*np.sin(np.pi*(i+1) *(K+.5)/N)

    coeff = 1/np.float_power( N,decay_rate )
    modes.append( p[-1]*coeff/normalizer)

    p_f2 +=(-1)**(K)/2*modes[-1]
    return p_f2

def _inverse_sin_discrete_transform_KLFull(p, N, var, tau, gamma):
    """Inverse of sin discrete transform, the code
    computes the KL expansion coefficients that the KLExpansion_Full geometry computes given p and then computes the same transformation done by `idst` from
    `scipy.fftpack`"""
    p_f2 = np.zeros(N)
    modes = []
    K = np.arange(N)
    for i in range(0,N-1):

        coeff = var*np.float_power(tau, gamma)*np.float_power( i**2+tau,-gamma)
        modes.append( p[i]*coeff)
        p_f2 +=  modes[-1]*np.sin(np.pi*(i+1) *(K+.5)/N)

    coeff = var*np.float_power(tau, gamma)*np.float_power( (N-1)**2+tau,-gamma)
    modes.append( p[-1]*coeff)

    p_f2 +=(-1)**(K)/2*modes[-1]

    return p_f2/np.pi

def test_KLExpansion_Full_geometry():
    """Check KLExpansion_Full geometry correctness"""
    N = 20
    grid = np.linspace(0,1,N)
    std = 1.2
    cor_len = .1
    nu = 3.0
    geom = cuqi.geometry.KLExpansion_Full(grid,std,cor_len,nu)

    p = np.random.randn(N)
    f_geom = geom.par2fun(p)
    var = std**2
    tau = 1.0/cor_len**2
    gamma = nu+1.
    f_expected = _inverse_sin_discrete_transform_KLFull(p, N, var, tau, gamma)

    assert np.allclose(f_geom, f_expected)

def test_create_CustomKL_geometry():
    """Check CustomKL geometry initialization"""
    N = 20
    grid = np.linspace(0,1,N)
    mean = 1
    std = .1
    cov_func = None
    trunc_term = 4
    geom = cuqi.geometry.CustomKL(grid,mean,std,cov_func,trunc_term)

    assert np.isclose(geom.mean, mean) and\
           np.isclose(geom.std, std) and\
	   geom.trunc_term==trunc_term
	

def test_KLExpansion_projection(copy_reference):
    """Check KLExpansion geometry projection performed by the method fun2par)"""
    # Set up a KLExpansion geometry
    num_modes = 95
    N = 100
    L = 1.0
    grid = np.linspace(0,1,N)

    geom = cuqi.geometry.KLExpansion(grid, num_modes=num_modes,
     decay_rate=1.5,
     normalizer=12.0)

    # Create a signal 
    signal =1/30*(1-np.cos(2*np.pi*(L-grid)/(L)))\
                    +1/30*np.exp(-2*(10*(grid-0.5))**2)+\
                     1/30*np.exp(-2*(10*(grid-0.8))**2)

    # Project signal to the KL basis and back
    p = geom.fun2par(signal)
    assert(len(p) == num_modes)

    signal_proj = geom.par2fun(p)
    assert(len(signal_proj) == N)

    # Check that the projection is accurate
    rel_err = np.linalg.norm(signal-signal_proj)/np.linalg.norm(signal)
    assert np.isclose(rel_err, 0.0, atol=1e-5)

    # Compare results with reference data
    ref_file = copy_reference("data/geometry/test_KLExpansion_projection.npz")
    ref = np.load(ref_file)
    assert np.allclose(p, ref["p"])

def test_DefaultGeometry2D_should_be_image2D():
    geom2D = cuqi.geometry._DefaultGeometry2D((100, 100))

    assert isinstance(geom2D, cuqi.geometry.Image2D)
   

@pytest.fixture
def geom2D_funvals():
    """Returns two different function values for a continuous 2D geometry"""
    geom2D = cuqi.geometry.Continuous2D((3, 3))
    def func1(x, y): return x*y
    def func2(x, y): return x**2*y**2
    XX, YY = np.meshgrid(geom2D.grid[0], geom2D.grid[1])
    funval1 = cuqi.array.CUQIarray(
        func1(XX, YY), is_par=False, geometry=geom2D)
    funval2 = cuqi.array.CUQIarray(
        func2(XX, YY), is_par=False, geometry=geom2D)
    return funval1, funval2


def test_Continuous2D_par2fun_and_fun2par_correctness(geom2D_funvals):
    """Check the correctness of the par2fun and fun2par methods for a continuous
    2D geometry"""
    funval1, funval2 = geom2D_funvals
    geom2D = funval1.geometry

    # Checks for one function value
    assert (funval1.shape == (3, 3))
    assert (funval1.parameters.shape == (9,))
    assert (funval1.parameters.funvals.shape == (3, 3))
    assert np.allclose(funval1, funval1.parameters.funvals)

    # Checks for an array of more than one function values
    # create the array
    multiple_funvals = np.stack((funval1, funval2), axis=-1)
    assert (multiple_funvals.shape == (3, 3, 2))

    # convert to parameters
    multiple_funvals_topar = geom2D.fun2par(multiple_funvals)
    assert np.allclose(multiple_funvals_topar[:, 0], funval1.parameters)
    assert np.allclose(multiple_funvals_topar[:, 1], funval2.parameters)

    # convert back to function values
    multiple_funvals_topar_tofun = geom2D.par2fun(multiple_funvals_topar)
    assert np.allclose(multiple_funvals_topar_tofun, multiple_funvals)


def test_Continuous2D_plot_multiple_funvals_pass(geom2D_funvals):
    """Check the correctness of the plot method for a continuous 2D geometry
    when multiple function values are given"""
    funval1, funval2 = geom2D_funvals
    geom2D = funval1.geometry
    multiple_funvals = np.stack((funval1, funval2), axis=-1)
    assert (multiple_funvals.shape == (3, 3, 2))
    # Test plotting passes
    geom2D.plot(multiple_funvals, is_par=False)
    # Convert to parameters then plot
    multiple_funvals_topar = geom2D.fun2par(multiple_funvals)
    geom2D.plot(multiple_funvals_topar, is_par=True)

# Create fixture for KLExpansion geometry
@pytest.fixture
def geom_KL():
    """Returns a KLExpansion geometry"""
    N = 20
    grid = np.linspace(0, 1, N)
    decay_rate = 2.5
    normalizer = 12
    geom = cuqi.geometry.KLExpansion(grid,
                                     decay_rate=decay_rate,
                                     normalizer=normalizer,
                                     num_modes=10)
    return geom

# Create fixture for StepExpansion geometry
@pytest.fixture
def geom_Step():
    """Returns a StepExpansion geometry"""
    N = 20
    grid = np.linspace(0, 1, N)
    n_steps = 10
    geom = cuqi.geometry.StepExpansion(grid, n_steps=n_steps)
    return geom

# Create fixture to parametrize the tests by geometry
@pytest.fixture
def geom(request):
    return request.getfixturevalue(request.param)

# Compare par2fun and fun2par for each geometry for individual parameters 
# vectors and multiple parameter vectors
@pytest.mark.parametrize('geom', ['geom_KL', 'geom_Step'], indirect=True)
def test_par2fun_and_fun2par_correctness_for_multiple_values(geom):
    """Check the correctness of the par2fun and fun2par methods for different
    geometries for multiple parameter vectors"""

    # Create random parameter values
    np.random.seed(0)
    n = 5
    par = np.random.randn(geom.par_dim, n)
    
    # run par2fun (first for multiple parameter vectors at once)
    fun = geom.par2fun(par)

    # run par2fun (for each parameter vector individually)
    fun_ind = np.array([geom.par2fun(par[:, i]) for i in range(n)]).T

    # Check that the results are the same
    assert np.allclose(fun, fun_ind)

    # Check plotting runs
    geom.plot(fun, is_par=False)

    # run fun2par (first for multiple functions at once)
    par2 = geom.fun2par(fun)

    # run fun2par (for each function individually)
    par2_ind = np.array([geom.fun2par(fun[:, i]) for i in range(n)]).T

    # Check that the results are the same
    assert np.allclose(par2, par2_ind)

    # Check plotting runs
    geom.plot(par2, is_par=True)



import cuqi
import numpy as np
import pytest
from .test_model import MultipleInputTestModel

def test_joint_dist_dim_geometry():
    """ Test the dimension and geometry properties of a joint distribution """

    model, data, _ = cuqi.testproblem.Poisson1D().get_components() # Model is m times n, m != n.

    # Bayesian model
    d = cuqi.distribution.Gamma(1, 1e-4)
    x = cuqi.distribution.Gaussian(np.zeros(model.domain_dim), lambda d: d)
    y = cuqi.distribution.Gaussian(model, 1)

    # Joint distribution
    J = cuqi.distribution.JointDistribution(d, x, y)

    # Test the dimensions
    assert J.dim == [d.dim, x.dim, y.dim]

    # Test the geometries
    assert J.geometry == [d.geometry, x.geometry, y.geometry]

    # Test dim and geometry for posterior
    P = J(y=data)

    assert P.dim == [d.dim, x.dim]

    assert P.geometry == [d.geometry, x.geometry]


@pytest.mark.parametrize("densities", [
    [
        cuqi.distribution.Gamma(1, 1e-4, name="x"),
        cuqi.distribution.Normal(0, 1, name="y")
    ],
    [
        cuqi.distribution.Gamma(1, 1e-4, name="x"),
        cuqi.distribution.Normal(0, lambda x:x, name="y")
    ],
    [
        cuqi.distribution.Gamma(1, 1e-4, name="d"),
        cuqi.distribution.Gamma(1, 1e-2, name="l"),
        cuqi.distribution.Gaussian(np.zeros(8), lambda d: d, name="x"),
        cuqi.distribution.Gaussian(
            mean=cuqi.testproblem.Deconvolution1D(dim=8).model,
            cov=lambda l: l,
            name="y"
        )
    ],
    [
        cuqi.distribution.Normal(0, 1e-2, name="z"),
        cuqi.distribution.Gamma(1, lambda z: abs(z), name="d"),
        cuqi.distribution.Gamma(lambda z: z, 1e-2, name="l"),
        cuqi.distribution.Gaussian(np.zeros(8), lambda d: d, name="x"),
        cuqi.distribution.Gaussian(
            mean=cuqi.testproblem.Deconvolution1D(dim=8).model,
            cov=lambda l: l,
            name="y"
        )
    ],
])
def test_joint_dist_logd(densities):
    """ Tests the log density function of the Joint distribution """

    # Create a joint distribution
    J = cuqi.distribution.JointDistribution(*densities)

    # Get the parameter names
    names = J.get_parameter_names()

    # Get dimensions
    dims = J.dim

    # Make dictionary of random parameters matching names and dimensions
    params = {name: abs(np.random.randn(dim)) for name, dim in zip(names, dims)}

    # Evaluate the log density function at random points
    logd = J.logd(**params)

    # Manually evaluate the log density function for each density
    logd_manual = 0
    for density in densities:
        # Extract the parameters for this density
        density_params = {name: params[name] for name in density.get_parameter_names()}
        logd_manual += density.logd(**density_params)

    # Compare the log density functions
    assert logd == pytest.approx(logd_manual)

@pytest.mark.parametrize("densities", [
    [
        cuqi.distribution.Gamma(1, 1e-4, name="x"),
        cuqi.distribution.Gaussian(0, lambda x:x, name="z"),
        cuqi.distribution.Normal(0, lambda x:x, name="y")
    ],
    [
        cuqi.distribution.Gamma(1, 1e-4, name="d"),
        cuqi.distribution.Gamma(1, 1e-2, name="l"),
        cuqi.distribution.Gaussian(np.zeros(8), lambda d: d, name="x"),
        cuqi.distribution.Gaussian(
            mean=cuqi.testproblem.Deconvolution1D(dim=8).model,
            cov=lambda l: l,
            name="y"
        )
    ],
    [
        cuqi.distribution.Normal(0, 1e-2, name="z"),
        cuqi.distribution.Gamma(1, lambda z: abs(z), name="d"),
        cuqi.distribution.Gamma(lambda z: z, 1e-2, name="l"),
        cuqi.distribution.Gaussian(np.zeros(8), lambda d: d, name="x"),
        cuqi.distribution.Gaussian(
            mean=cuqi.testproblem.Deconvolution1D(dim=8).model,
            cov=lambda l: l,
            name="y"
        )
    ],
])
def test_joint_dist_properties(densities):
    """ This tests various properties of the joint distribution """

    # Create a joint distribution
    J = cuqi.distribution.JointDistribution(*densities)

    # Now check various properties. First check the dimension
    assert J.dim == [density.dim for density in densities]

    # Check the geometry
    assert J.geometry == [density.geometry for density in densities]

    # Check the parameter names
    assert J.get_parameter_names() == [density.name for density in densities]

    # Check list of distributions
    assert J._distributions == densities

    # Check likelihoods
    assert J._likelihoods == []

    # Now we condition y on some data
    data = cuqi.testproblem.Deconvolution1D(dim=8).data

    P = J(y=data)

    # Check the dimension
    assert P.dim == [density.dim for density in densities[:-1]]

    # Check the geometry
    assert P.geometry == [density.geometry for density in densities[:-1]]

    # Check the parameter names
    assert P.get_parameter_names() == [density.name for density in densities[:-1]]

    # Check list of distributions by comparing the names
    assert [dist.name for dist in P._distributions] == [density.name for density in densities[:-1]]

    # Check likelihoods by comparing the names
    assert [L.name for L in P._likelihoods] == [densities[-1].name]

def test_joint_dist_reduce():
    """ This tests the reduce hack for the joint distribution. """

    J, data = hierarchical_joint()   

    # Check if we get the expected result when conditioning.
    assert isinstance(J(y=data), cuqi.distribution.JointDistribution)
    assert isinstance(J(y=data, z=1, d=1, l=1), cuqi.distribution.Posterior)
    assert isinstance(J(y=data, z=1, d=1, x=np.zeros(8)), cuqi.distribution.Posterior)
    assert isinstance(J(y=data, z=1, x=np.zeros(8), l=1), cuqi.distribution.Posterior)
    assert isinstance(J(y=data, d=1, x=np.zeros(8), l=1), cuqi.distribution.JointDistribution) # 2 likelihoods

def test_stacked_joint_distribution():
    """ This tests the stacked joint distribution """

    J, data = hierarchical_joint()

    Js = J._as_stacked()

    # Check the dimension
    assert Js.dim == sum(J.dim)

    # Check the geometry
    assert Js.geometry == cuqi.geometry._DefaultGeometry1D(Js.dim)

    # Check you can evaluate the log density function with a single vector
    Js.logd(np.ones(Js.dim))

    # Check that you can condition on the stacked distribution
    # and again evaluate as a single vector
    Ps = Js(y=data)
    Ps.logd(np.ones(Ps.dim))

    # Even if you condition in multiple parameters
    Ps2 = Js(y=data, z=1, d=1)
    Ps2.logd(np.ones(Ps2.dim))

    # And even if you condition on all parameters
    Ps3 = Js(y=data, z=1, d=1, x=np.zeros(Js.get_density("x").dim), l=1)
    #Ps3.logd() # Should be allowed with no arguments

def hierarchical_joint(main_dim=8):
    """ This creates a deconvolution based hierarchical joint distribution.
    
    Parameters
    ----------
    main_dim : int
        The dimension of the main model.

    Returns
    -------
    JointDistribution, data
        
    """

    densities = [
        cuqi.distribution.Normal(0, 1e-2, name="z"),
        cuqi.distribution.Gamma(1, lambda z: abs(z), name="d"),
        cuqi.distribution.Gamma(lambda z: z, 1e-2, name="l"),
        cuqi.distribution.Gaussian(np.zeros(main_dim), lambda d: d, name="x"),
        cuqi.distribution.Gaussian(
            mean=cuqi.testproblem.Deconvolution1D(dim=main_dim).model,
            cov=lambda l: l,
            name="y"
        )
    ]

    data = cuqi.testproblem.Deconvolution1D(dim=main_dim).data

    J = cuqi.distribution.JointDistribution(*densities)

    return J, data

def test_extra_parameter_no_prior():
    A = np.random.randn(10,3)

    # Define distributions for Bayesian model
    y = cuqi.distribution.Normal(lambda x: A@x, np.ones(10))
    x = cuqi.distribution.Normal(np.zeros(3), lambda z,b:z+b)
    z = cuqi.distribution.Gamma(1, 1)

    # Joint distribution p(y,x,z)
    with pytest.raises(ValueError, match=r"Missing prior for \['b'\]"):
        cuqi.distribution.JointDistribution(y,x,z)


def test_JointDistribution_reduce_MultipleLikelihoodPosterior():
    """ This tests if the joint distribution can be reduced to MultipleLikelihoodPosterior """

    J, data = hierarchical_joint()

    # Posterior w.r.t z is with multiple likelihoods if all other parameters are fixed
    # for this J
    posterior_z = J(y=data, d=0.1, l=0.1, x=data)

    # Check we get the right class
    assert isinstance(posterior_z, cuqi.distribution.MultipleLikelihoodPosterior)

    # Check the dimension
    assert posterior_z.dim == 1

    # Check the "prior" in multiple likelihood posterior matches z
    assert posterior_z.prior.mean == J.get_density("z").mean
    assert posterior_z.prior.std == J.get_density("z").std

    # Check logd can be evaluated (tolerance is high since new data every time)
    assert np.allclose(posterior_z.logd(1), -5000, atol=10)

def test_MultipleLikelihoodPosterior_should_raise_if_two_densities():
    """ This tests if the MultipleLikelihoodPosterior raises if it has two densities. """

    # Define distributions for Bayesian model
    y = cuqi.distribution.Normal(lambda x: x, 1)
    x = cuqi.distribution.Normal(0, 1)

    # Joint distribution p(y,x)
    J = cuqi.distribution.JointDistribution(y,x)

    # Posterior
    P = J(y=1)

    # Check we get the right class
    assert isinstance(P, cuqi.distribution.Posterior)

    # Check that we cannot create MultipleLikelihoodPosterior
    with pytest.raises(ValueError, match=r"requires at least three densities"):
        cuqi.distribution.MultipleLikelihoodPosterior(y.to_likelihood(1), x)


def test_MultipleLikelihoodPosterior_should_raise_if_no_likelihood():
    """ This tests if the MultipleLikelihoodPosterior raises if no likelihood is given."""

    # Define distributions for Bayesian model
    y1 = cuqi.distribution.Normal(lambda x: x, 1)
    y2 = cuqi.distribution.Normal(lambda x: x+1, 1)
    x = cuqi.distribution.Normal(0, 1)
 
    # Check that we cannot create MultipleLikelihoodPosterior
    with pytest.raises(ValueError, match=r"must have a likelihood and prior"):
        cuqi.distribution.MultipleLikelihoodPosterior(y1, y2, x)

def test_MultipleLikelihoodPosterior_should_raise_if_names_do_not_match():
    """ This tests if the MultipleLikelihoodPosterior raises if the names do not match."""

    # Define distributions for Bayesian model
    y = cuqi.distribution.Normal(lambda x: x, 1)
    x = cuqi.distribution.Normal(0, 1)
    z = cuqi.distribution.Normal(0, 1)
 
    # Check that we cannot create MultipleLikelihoodPosterior
    with pytest.raises(ValueError, match=r"the same parameter name"):
        cuqi.distribution.MultipleLikelihoodPosterior(y.to_likelihood(1), x, z)

@pytest.mark.parametrize("joint, variables", [
    (
        cuqi.distribution.JointDistribution(
            cuqi.distribution.Normal(0, 1, name="x"),
            cuqi.distribution.Normal(0, 1, name="y")
        ),
        {
            "x": 0,
            "y": 0
        }
    ),
    (
        cuqi.distribution.JointDistribution(
            cuqi.distribution.Uniform(0, 10, name="d"),
            cuqi.distribution.Uniform(0, 5, name="s"),
            cuqi.distribution.Gaussian(np.zeros(8), lambda d: d, name="x"),
            cuqi.distribution.Gaussian(
                mean=cuqi.testproblem.Deconvolution1D(dim=8).model,
                cov=lambda s: s,
                name="y"
            )
        ),
        {
            "d": 5,
            "s": 2,
            "x": np.zeros(8),
            "y": cuqi.testproblem.Deconvolution1D(dim=8).data
        }
    ),
    (
        cuqi.distribution.JointDistribution(
            cuqi.distribution.Uniform(0, 10, name="d"),
            cuqi.distribution.Uniform(0, 5, name="s"),
            cuqi.distribution.Gaussian(np.zeros(8), lambda d: d, name="x"),
            cuqi.distribution.Gaussian(
                mean=cuqi.testproblem.Deconvolution1D(dim=8).model,
                cov=lambda s: s,
                name="y"
            )
        ),
        {
            "d": 11, # Out of bounds (all should return -Inf)
            "s": 2,
            "x": np.zeros(8),
            "y": cuqi.testproblem.Deconvolution1D(dim=8).data
        }
    )
    ]
)
def test_logd_consistency_when_conditioning(joint, variables):
    """ Test consistency of logd value when conditioning the joint distribution.
     
    This ensures we always return the correct value for logd even when reducing to single density.
    
    """

    # True value of logd by fully evaluating the joint distribution
    true_value = joint.logd(**variables)

    # Loop over all variables and evaluate the logd value
    # where all previously seen variables are conditioned
    # and not seen variables are given to logd.
    cond_vars = {}
    for key, value in variables.items():
        # Remaining variables are those that are not used for conditioning
        remaining_vars = {k: v for k, v in variables.items() if k not in cond_vars}
        
        # Condition the joint distribution
        cond_joint = joint(**cond_vars)

        # Evaluate the logd value of conditioned joint distribution
        # This may be a single density or a joint distribution since
        # joint can reduce to single density (such as Posterior)
        logd_value = cond_joint.logd(**remaining_vars)

        # Potential error message if this assert fails:
        msg = f"Evaluated: \n {joint} \n with variables {variables}.\nFailed at {key}={value} for \n {cond_joint} \n with variables {remaining_vars}."
        
        # Assert the logd value matches
        assert logd_value == pytest.approx(true_value, rel=1e-6), msg

        # Add current variable to the variables that need to be conditioned
        cond_vars[key] = value

def test_joint_distribution_with_multiple_inputs_model_has_correct_parameter_names():
    """Test that the joint distribution based on model with multiple inputs has
    correct parameter names."""

    test_model = MultipleInputTestModel.helper_build_three_input_test_model()
    model = cuqi.model.Model(
        test_model.forward_map,
        gradient=test_model.gradient_form2,
        domain_geometry=test_model.domain_geometry,
        range_geometry=test_model.range_geometry,
    )

    # Create priors
    x_dist = cuqi.distribution.Gaussian(
        mean=np.zeros(3),
        cov=np.eye(3))
    y_dist = cuqi.distribution.Gaussian(
        mean=np.zeros(2),
        cov=np.eye(2))
    z_dist = cuqi.distribution.Gaussian(
        mean=np.zeros(3),
        cov=np.eye(3))

    # Create data distribution
    data_dist = cuqi.distribution.Gaussian(
        mean=model(x_dist, y_dist, z_dist), cov = 1.0)

    # Create likelihood
    likelihood = data_dist(data_dist = np.array([2,2,3]))

    x_val = np.array([1, 2, 3])
    y_val = np.array([4, 5])
    z_val = np.array([6, 7, 8])

    posterior = cuqi.distribution.JointDistribution(
        likelihood,
        x_dist,
        y_dist,
        z_dist
    )

    # Ensure correct parameter names are returned for joint distribution with likelihood
    assert posterior.get_parameter_names() == ['x_dist', 'y_dist', 'z_dist']

    assert posterior(x_dist=x_val).get_parameter_names() == ['y_dist', 'z_dist']
    assert posterior(y_dist=y_val).get_parameter_names() == ['x_dist', 'z_dist']
    assert posterior(z_dist=z_val).get_parameter_names() == ['x_dist', 'y_dist']

    assert posterior(y_dist=y_val, z_dist=z_val).get_parameter_names() == ['x_dist']
    assert posterior(x_dist=x_val, z_dist=z_val).get_parameter_names() == ['y_dist']
    assert posterior(x_dist=x_val, y_dist=y_val).get_parameter_names() == ['z_dist']

    assert posterior(x_dist=x_val, y_dist=y_val, z_dist=z_val).get_parameter_names() == []

    joint_dist =  cuqi.distribution.JointDistribution(
        data_dist,
        x_dist,
        y_dist,
        z_dist
    )

    # Ensure correct parameter names are returned for joint distribution with data distribution
    assert joint_dist.get_parameter_names() == ['data_dist', 'x_dist', 'y_dist', 'z_dist']
    assert joint_dist(x_dist=x_val).get_parameter_names() == ['data_dist', 'y_dist', 'z_dist']
    assert joint_dist(y_dist=y_val).get_parameter_names() == ['data_dist', 'x_dist', 'z_dist']
    assert joint_dist(z_dist=z_val).get_parameter_names() == ['data_dist', 'x_dist', 'y_dist']
    assert joint_dist(data_dist=np.array([2,2,3])).get_parameter_names() == ['x_dist', 'y_dist', 'z_dist']

    assert joint_dist(x_dist=x_val, data_dist=np.array([2,2,3])).get_parameter_names() == ['y_dist', 'z_dist']
    assert joint_dist(y_dist=y_val, data_dist=np.array([2,2,3])).get_parameter_names() == ['x_dist', 'z_dist']
    assert joint_dist(z_dist=z_val, data_dist=np.array([2,2,3])).get_parameter_names() == ['x_dist', 'y_dist']
    assert joint_dist(x_dist=x_val, y_dist=y_val).get_parameter_names() == ['data_dist', 'z_dist']
    assert joint_dist(x_dist=x_val, z_dist=z_val).get_parameter_names() == ['data_dist', 'y_dist']
    assert joint_dist(y_dist=y_val, z_dist=z_val).get_parameter_names() == ['data_dist', 'x_dist']

    assert joint_dist(x_dist=x_val, y_dist=y_val, z_dist=z_val).get_parameter_names() == ['data_dist']
    assert joint_dist(x_dist=x_val, y_dist=y_val, data_dist=np.array([2,2,3])).get_parameter_names() == ['z_dist']
    assert joint_dist(x_dist=x_val, z_dist=z_val, data_dist=np.array([2,2,3])).get_parameter_names() == ['y_dist']
    assert joint_dist(y_dist=y_val, z_dist=z_val, data_dist=np.array([2,2,3])).get_parameter_names() == ['x_dist']

    # Ensure correct parameter names are returned for underlying likelihood
    assert joint_dist(data_dist=np.array([2,2,3]))._likelihoods[0].get_parameter_names() == ['x_dist', 'y_dist', 'z_dist']

    assert joint_dist(x_dist=x_val, data_dist=np.array([2,2,3]))._likelihoods[0].get_parameter_names() == ['y_dist', 'z_dist']
    assert joint_dist(y_dist=y_val, data_dist=np.array([2,2,3]))._likelihoods[0].get_parameter_names() == ['x_dist', 'z_dist']
    assert joint_dist(z_dist=z_val, data_dist=np.array([2,2,3]))._likelihoods[0].get_parameter_names() == ['x_dist', 'y_dist']

    assert joint_dist(x_dist=x_val, y_dist=y_val, data_dist=np.array([2,2,3])).likelihood.get_parameter_names() == ['z_dist']
    assert joint_dist(x_dist=x_val, z_dist=z_val, data_dist=np.array([2,2,3])).likelihood.get_parameter_names() == ['y_dist']
    assert joint_dist(y_dist=y_val, z_dist=z_val, data_dist=np.array([2,2,3])).likelihood.get_parameter_names() == ['x_dist']


import cuqi
import numpy as np

import sys

from cuqi.distribution import Gaussian, CMRF, Gaussian, LMRF, GMRF
from cuqi.sampler import pCN

import pytest


def test_CWMH_modify_proposal():
    # Parameters
    n = 2
    scale = 0.05*np.ones(n)
    x0 = 0.5*np.ones(n)

    # Set up target
    target_dist = cuqi.distribution.CMRF(np.zeros(n), 0.5, 'neumann')
    def target(x): return target_dist.pdf(x)

    # Set up proposals
    #proposal1 = cuqi.distribution.Normal(np.zeros(n),1 )
    #proposal2 = cuqi.distribution.Normal(np.zeros(n),2 )
    #def proposal1(x, sigma): return np.random.normal(x, sigma)
    #def proposal2(x, sigma): return np.random.normal(x, 2*sigma)

    proposal1 =cuqi.distribution.Normal(mean = lambda location:location,std = lambda scale:scale, geometry=n)
    proposal2 =cuqi.distribution.Normal(mean = lambda location:location,std = lambda scale:2*scale, geometry=n)


    # Set up sampler
    MCMC1 = cuqi.sampler.CWMH(target, proposal1, scale, x0)
    MCMC2 = cuqi.sampler.CWMH(target, proposal2, scale, x0)

    # Switch proposal
    MCMC1.proposal = proposal2

    # Compute samples 1
    np.random.seed(0)
    results1 = MCMC1.sample(10,2).samples

    # Compute samples 2
    np.random.seed(0)
    results2 = MCMC2.sample(10,2).samples

    # Compare samples
    assert np.allclose(results1,results2)

def test_CWMH_sample_regression():
    # Set seed
    np.random.seed(0)
    
    # Define target
    mean = np.array([0, 0])
    std = np.array([1, 1])
    R = np.array([[1, -0.7], [-0.7, 1]])
    cov = np.diag(std) @ (R @ np.diag(std))
    def target(x): return cuqi.distribution.Gaussian(mean,cov).pdf(x)

    # Define proposal
    # def proposal(x, sigma): return np.random.normal(x, sigma)
    #proposal = cuqi.distribution.Normal(np.zeros(len(mean)),1 )
    proposal =cuqi.distribution.Normal(mean = lambda location:location,std = lambda scale:scale, geometry=2)

    # Set up sampler
    MCMC = cuqi.sampler.CWMH(target, proposal, 0.05, np.array([0,0]))

    # Compare with previously computed results
    np.allclose(MCMC.sample(5,1).samples,np.array([[ 0.18158052,  0.17641957,  0.21447146,  0.23666462,  0.23666462],[-0.02885603, -0.00832611, -0.00224236,  0.01444136,  0.01444136]]))


def test_RWMH_sample_regression():
    # Set seed
    np.random.seed(0) #nice seed!

    d = 6
    mu = np.zeros(d)
    sigma = np.linspace(0.5, 1, d)

    # target function to sample
    dist = cuqi.distribution.Gaussian(mu, sigma**2)

    ref = cuqi.distribution.Gaussian(mu, np.ones(d))   # standard Gaussian

    scale = 0.1
    x0 = 0.5*np.ones(d)
    MCMC2 = cuqi.sampler.MH( dist,proposal = ref,scale =scale, x0=x0)

    # run sampler
    Ns = int(1e1)      # number of samples
    Nb = int(0.2*Ns)   # burn-in

    #
    x_s2 = MCMC2.sample_adapt(Ns, Nb)
    reference = np.array([[ 0.5       ,  0.77238519,  0.73381779,  0.7700134 ,  0.41274389,
         0.18348216,  0.37057737,  0.34837564,  0.34837564,  0.34837564],
       [ 0.5       ,  0.46259765,  0.70379628,  0.79213478,  1.00263215,
         0.80556899,  0.33926608,  0.67670237,  0.67670237,  0.67670237],
       [ 0.5       ,  0.5570753 ,  0.50402891,  0.29664278,  0.16967913,
        -0.39387781, -0.57857024, -0.52616853, -0.52616853, -0.52616853],
       [ 0.5       ,  0.34429001,  0.35183386, -0.1108788 , -0.366134  ,
        -0.01638806, -0.23662861,  0.00165624,  0.00165624,  0.00165624],
       [ 0.5       ,  0.03456505,  0.2675372 ,  0.18626517,  0.51389191,
        -0.21210323,  0.46072045, -0.03899323, -0.03899323, -0.03899323],
       [ 0.5       ,  0.61916105,  0.86364669,  0.9001697 ,  0.26407212,
         0.16837312, -0.10678787, -0.39255235, -0.39255235, -0.39255235]])

    assert np.allclose(x_s2.samples,reference)

def test_pCN_sample_regression():
    np.random.seed(0)
    d= 2
    mu = np.zeros(d)
    sigma = np.linspace(0.5, 1, d)
    model = cuqi.model.Model(lambda x: x, range_geometry=d, domain_geometry=d)
    L = Gaussian(mean=model, sqrtcov=sigma).to_likelihood(np.zeros(d))
    def target(x): return L.logd(x)
    P = Gaussian(mu, np.ones(d))
    scale = 0.1
    x0 = 0.5*np.ones(d)
    posterior = cuqi.distribution.Posterior(L, P)
    MCMC = pCN(posterior, scale, x0)
    results = MCMC.sample(10,2)
    assert np.allclose(results.samples, np.array([[0.44368817, 0.44368817, 0.56807601, 0.64133227, 0.64133227,
         0.78752546, 0.68750247, 0.42875734, 0.40239322, 0.40495205],
        [0.66816968, 0.66816968, 0.61423278, 0.6233214 , 0.6233214 ,
         0.59968114, 0.6343679 , 0.69654995, 0.84449757, 0.82154609]]))

    assert np.allclose(results.loglike_eval, np.array([-1.76167363, -1.76167363, -1.97879155, -2.16160882, -2.16160882,
        -2.5649313 , -2.2912605 , -1.75498652, -1.82515857, -1.8101712 ]))

    assert results.acc_rate==0.8


def test_sampler_geometry_assignment():

    n = 2
    scale = 0.05*np.ones(n)
    x0 = 0.5*np.ones(n)

    # Set up target
    target = cuqi.distribution.CMRF(np.zeros(n), 0.5, 'neumann')
    target.geometry = cuqi.geometry.Continuous2D((1,2))

    # Set up proposals
    proposal =cuqi.distribution.Normal(mean = lambda location:location,std = lambda scale:scale, geometry=n)

    # Set up sampler
    MCMC_sampler = cuqi.sampler.CWMH(target, proposal, scale, x0)
    samples = MCMC_sampler.sample(10,2)

    assert(MCMC_sampler.geometry == target.geometry and\
           MCMC_sampler.geometry == samples.geometry)

def test_sampler_UserDefined_basic():
    """
    This tests all samplers which can work with any logpdf as input 
    (potentially requireing a gradient or sample method also)
    """
    # Distribution
    X = cuqi.distribution.Gaussian(np.array([2,3]),np.array([[2,0.1],[0.1,5]]))
    distX = cuqi.distribution.UserDefinedDistribution(X.dim, X.logpdf, X.gradient, X.sample)

    # Parameters
    Ns = 500   # number of samples
    Nb = 100   # burn-in

    s_MH = cuqi.sampler.MH(distX).sample_adapt(Ns,Nb)
    s_CWMH = cuqi.sampler.CWMH(distX).sample_adapt(Ns,Nb)
    s_NUTS = cuqi.sampler.NUTS(distX).sample_adapt(Ns,Nb)

    assert np.allclose(s_MH.shape,(X.dim,Ns))
    assert np.allclose(s_CWMH.shape,(X.dim,Ns))
    assert np.allclose(s_NUTS.shape,(X.dim,Ns))

def test_sampler_UserDefined_tuple():
    """
    This tests samplers with a two userdefined distributions.
    Here we require two userdefined distributions.
    """
    # This provides a way to give the logpdf
    P = cuqi.distribution.Gaussian(np.array([2,3]),np.array([[2,0.1],[0.1,5]]))

    #
    model = cuqi.model.Model(lambda x: x, range_geometry=2, domain_geometry=2)
    L = cuqi.distribution.Gaussian(model,np.array([[1,0.5],[0.5,3]])).to_likelihood(np.array([5,6]))

    # Define userdefined distribution + likelihood
    userP = cuqi.distribution.UserDefinedDistribution(2, P.logpdf, P.gradient, P.sample)

    userL = cuqi.likelihood.UserDefinedLikelihood(2, L.logd, L.gradient)

    # Parameters
    Ns = 2000   # number of samples
    Nb = 200   # burn-in

    # Run samplers
    s_pCN = cuqi.sampler.pCN((userL,userP)).sample_adapt(Ns)

    assert np.allclose(s_pCN.shape,(P.dim,Ns))

def test_sampler_CustomInput_LinearRTO():
    """
    This tests the LinearRTO sampler which requires a specific input structure.
    """
    # In some special cases A model must be part of the likelihood (e.g. in LinearRTO)
    model = np.eye(2) # Identity model
    #model = cuqi.model.LinearModel(lambda x: x, lambda y: y, 2, 2) 

    # In LinearRTO we require Gaussian distributions
    # or at least classes with sqrtprec and sqrtprecTimesMean 
    # This is most easily done by defining a Gaussian. #TODO make UserDefined.
    P = cuqi.distribution.Gaussian(np.array([2,3]),np.array([[2,0.1],[0.1,5]]))
    L = cuqi.distribution.Gaussian(model,np.array([[1,0.5],[0.5,3]]))

    # Data
    data = np.array([5,6])

    # Posterior
    target = (data, model, L.sqrtprec, P.mean, P.sqrtprec)
    #target = cuqi.distribution.Posterior(L, P, data)

    # Parameters
    Ns = 2000   # number of samples
    Nb = 200   # burn-in

    # Sampling
    s_RTO = cuqi.sampler.LinearRTO(target).sample_adapt(Ns,Nb)

    assert np.allclose(s_RTO.shape,(P.dim,Ns))

def test_sampler_scalar_mean_Gaussian_LinearRTO():

    model = np.eye(2) # Identity model

    P = cuqi.distribution.Gaussian(0, 1, geometry=2)
    L = cuqi.distribution.Gaussian(model,np.array([[1,0.5],[0.5,3]]))

    # Data
    data = np.array([5,6])

    # Posterior
    target = (data, model, L.sqrtprec, P.mean, P.sqrtprec)

    # Parameters
    Ns = 200   # number of samples
    Nb = 20   # burn-in

    # Sampling
    s_RTO = cuqi.sampler.LinearRTO(target).sample_adapt(Ns,Nb)

    assert np.allclose(s_RTO.shape,(P.dim,Ns))


def test_ULA_UserDefinedDistribution():
    expected_samples = \
        np.array([[0.1, 0.11763052, 0.12740614],
                  [1.1, 1.10399157, 1.1263901 ]])
    np.random.seed(0)
    # Parameters
    dim = 2 # Dimension of distribution
    mu = np.arange(dim) # Mean of Gaussian
    std = 1 # standard deviation of Gaussian

    # Logpdf function
    logpdf_func = lambda x: -1/(std**2)*np.sum((x-mu)**2)
    gradient_func = lambda x: -2/(std**2)*(x - mu)

    # Define distribution from logpdf as UserDefinedDistribution (sample and gradients also supported)
    target = cuqi.distribution.UserDefinedDistribution(dim=dim, logpdf_func=logpdf_func, gradient_func=gradient_func)

    # Set up sampler
    sampler = cuqi.sampler.ULA(target, scale=.0001, x0=np.array([.1, 1.1]))

    # Sample
    samples = sampler.sample(3)

    assert np.allclose(samples.samples, expected_samples) and np.isclose(samples.acc_rate, 1)


def test_ULA_regression(copy_reference):
    # This tests compares ULA class results with results 
    # generate from original ULA code provided by Felipe Uribe.
    # The original code is found in commit:
    # c172442d8d7f34a33681b9c1d76889c99ac8dfcd

    np.random.seed(0)
    # %% Create CUQI test problem
    test = cuqi.testproblem._Deblur()
    n = test.model.domain_dim
    h = test.meshsize
    
    # Extract data
    data = test.data
    
    # Extract Likelihood
    likelihood  = test.likelihood
    
    # Define Prior
    loc = np.zeros(n)
    delta = 1
    scale = delta*h
    prior = cuqi.distribution.CMRF(loc, scale, 'neumann')
    
    # %% Create the posterior and the sampler
    posterior = cuqi.distribution.Posterior(likelihood, prior)
    MCMC = cuqi.sampler.ULA(posterior, scale=0.0001)

    # %% Sample
    samples  = MCMC.sample(5)
    samples_orig_file = copy_reference("data/ULA_felipe_original_code_results.npz")
    samples_orig = np.load(samples_orig_file)

    assert(np.allclose(samples.samples, samples_orig['arr_0']))


def test_MALA_UserDefinedDistribution():
    expected_samples = \
        np.array([[0.1, 0.11763052, 0.09493548],
                  [1.1, 1.10399157, 1.11731663]])
    np.random.seed(0)

    # Parameters
    dim = 2  # Dimension of distribution
    mu = np.arange(dim)  # Mean of Gaussian
    std = 1  # standard deviation of Gaussian

    # Logpdf function
    logpdf_func = lambda x: -1/(std**2)*np.sum((x-mu)**2)
    gradient_func = lambda x: -2/(std**2)*(x - mu)

    # Define distribution from logpdf as UserDefinedDistribution (sample and gradients also supported)
    target = cuqi.distribution.UserDefinedDistribution(
        dim=dim, logpdf_func=logpdf_func, gradient_func=gradient_func)

    # Set up sampler
    sampler = cuqi.sampler.MALA(target, scale=.0001, x0=np.array([.1, 1.1]))

    # Sample
    samples = sampler.sample(3)

    assert np.allclose(samples.samples, expected_samples)\
                    and np.isclose(samples.acc_rate, 1)


def test_MALA_regression(copy_reference):
    #%% CUQI
    dim = 5  # Dimension of distribution
    mu = np.arange(dim)  # Mean of Gaussian
    std = 1  # standard deviation of Gaussian
    N = 2000
    Nb = 500
    eps = 1/dim

    # Logpdf function
    logpdf_func = lambda x: -1/(std**2)*np.sum((x-mu)**2)
    gradient_func = lambda x: -2/(std**2)*(x-mu)
    
    # Define distribution from logpdf as UserDefinedDistribution (sample and gradients also supported)
    target = cuqi.distribution.UserDefinedDistribution(dim=dim, logpdf_func=logpdf_func,
                                                       gradient_func=gradient_func)

    # Set up sampler
    x0 = np.zeros(dim)
    np.random.seed(0)
    sampler = cuqi.sampler.MALA(target, scale=eps**2, x0=x0)
    # Sample
    samples = sampler.sample(N, Nb)

    samples_orig_file = copy_reference(
        "data/MALA_felipe_original_code_results.npz")
    samples_orig = np.load(samples_orig_file)

    assert(np.allclose(samples.samples, samples_orig['arr_0']))

@pytest.mark.parametrize("prior, sample_method, expected", [
    (Gaussian(np.zeros(128), 0.1), "_sampleMapCholesky", np.arange(10)), # Direct (no burn-in, no initial guess)
    (Gaussian(np.zeros(128), 0.1), "_sampleLinearRTO", np.arange(1,12)), # 20% burn-in + initial guess
    (Gaussian(np.zeros(128), 0.1), "_sampleNUTS", np.arange(1,12)),      # 20% burn-in + initial guess
    (Gaussian(np.zeros(128), 0.1), "_samplepCN", np.arange(1,12)),       # 20% burn-in + initial guess
    (Gaussian(np.zeros(128), 0.1), "_sampleCWMH", np.arange(1,12)),      # 20% burn-in + initial guess
    (LMRF(0, 0.1, geometry=128),"_sampleUGLA", np.arange(1,12)),   # 20% burn-in + initial guess
    ])
def test_TP_callback(prior, sample_method, expected):
    """ Test that the callback function is called with the correct sample index by comparing to the expected output.
    
    This tests the pipeline from testproblem all the way to the sampler.
    """

    TP = cuqi.testproblem.Deconvolution1D(dim=prior.dim)
    TP.prior = prior

    # Define callback that stores a list of sample indices
    Ns_list = []
    def callback(_, n):
        Ns_list.append(n)

    # Sampler
    sample_method_handle = getattr(TP, sample_method)
    
    # if sampler is MapCholesky, we do not pass burn-in
    # otherwise we pass burn-in Nb
    if sample_method == "_sampleMapCholesky":
        sample_method_handle(10, callback=callback)
    else:
        sample_method_handle(10, Nb=2, callback=callback)

    assert np.array_equal(Ns_list, expected)

def test_NUTS_regression(copy_reference):
    # SKIP NUTS test if not windows (for now) #TODO.
    if  not sys.platform.startswith('win') and \
       not sys.platform.startswith('darwin'):
        pytest.skip("NUTS regression test is not implemented for this platform")

    np.random.seed(0)
    tp = cuqi.testproblem.WangCubic()
    x0 = np.ones(tp.model.domain_dim)
    Ns = int(1e3)
    Nb = int(0.5*Ns)
    MCMC = cuqi.sampler.NUTS(tp.posterior, x0, max_depth = 12)
    samples = MCMC.sample(Ns, Nb)

    if sys.platform.startswith('win'):
        samples_orig_file = copy_reference(
            "data/NUTS_felipe_original_code_results_win.npz")
    elif sys.platform.startswith('darwin'):
        samples_orig_file = copy_reference(
            "data/NUTS_felipe_original_code_results_darwin.npz")
    
    samples_orig = np.load(samples_orig_file)

    assert(np.allclose(samples.samples, samples_orig["arr_0"]))

def _Gibbs_joint_hier_model(use_legacy=False, noise_std=0.01):
    """ Define a Gibbs sampler based on a joint distribution from a hierarchical model. Used for testing Gibbs sampler. """
    np.random.seed(0)
    
    # Model and data
    A, y_obs, _ = cuqi.testproblem.Deconvolution1D(phantom='square', use_legacy=use_legacy, noise_std=noise_std).get_components()
    n = A.domain_dim

    # Define distributions
    d = cuqi.distribution.Gamma(1, 1e-4)
    l = cuqi.distribution.Gamma(1, 1e-4)
    x = cuqi.distribution.GMRF(np.zeros(n), lambda d: d[0])
    y = cuqi.distribution.Gaussian(A, lambda l: 1/l)

    # Combine into a joint distribution and create posterior
    joint = cuqi.distribution.JointDistribution(d, l, x, y)
    posterior = joint(y=y_obs)

    # Define sampling strategy
    sampling_strategy = {
        'x': cuqi.sampler.LinearRTO,
        ('d', 'l'): cuqi.sampler.Conjugate,
    }

    # Define Gibbs sampler
    sampler = cuqi.sampler.Gibbs(posterior, sampling_strategy)

    return sampler

def test_Gibbs_regression(copy_reference):

    # SKIP Gibbs reg test if not windows.
    if not sys.platform.startswith('win'):
        pytest.skip("NUTS regression test is not implemented for this platform")

    # Legacy deconvolution test problem to match reference results
    sampler = _Gibbs_joint_hier_model(use_legacy=True, noise_std=0.05)

    # Run sampler
    samples = sampler.sample(Ns=100, Nb=20)

    if sys.platform.startswith('win'):
        samples_orig_file = copy_reference(
            "data/Gibbs_original_code_results_win.npz")
    samples_orig = np.load(samples_orig_file)

    # Save results
    #np.savez("Gibbs_original_code_results_win.npz", samples["d"].samples)

    assert(np.allclose(samples["d"].samples, samples_orig["arr_0"]))

def test_Gibbs_continue_sampling():
    """ This tests the sampling can continue with the Gibbs sampler """

    sampler = _Gibbs_joint_hier_model()

    # Run sampler
    samples = sampler.sample(Ns=10, Nb=5)

    # Continue sampling
    samples2 = sampler.sample(Ns=10)

    assert samples["x"].shape[-1] == 10
    assert samples2["x"].shape[-1] == 20

def test_Gibbs_geometry_matches():
    sampler = _Gibbs_joint_hier_model()

    target = sampler.target

    # Run sampler
    samples = sampler.sample(Ns=10, Nb=5)

    # Check that the geometry matches
    assert samples["d"].geometry == target.get_density("d").geometry
    assert samples["l"].geometry == target.get_density("l").geometry
    assert samples["x"].geometry == target.get_density("x").geometry

def test_RTO_with_AffineModel_is_equivalent_to_LinearModel_and_shifted_data():
    # Define LinearModel and data
    A, y_obs, _ = cuqi.testproblem.Deconvolution1D().get_components()

    # Define Shift
    shift = np.random.rand(A.domain_dim)

    # Define Bayesian Problem
    x = cuqi.distribution.GMRF(np.zeros(A.domain_dim), 100)
    y = cuqi.distribution.Gaussian(A@x, 0.01**2)
    posterior = cuqi.distribution.JointDistribution(x, y)(y=y_obs-shift)

    # Set up LinearRTO with both models
    sampler_linear = cuqi.sampler.LinearRTO(posterior)

    # Sample with fixes seed
    np.random.seed(0)
    samples_linear = sampler_linear.sample(10, 2)

    # Define AffineModel
    affine_model = cuqi.model.AffineModel(A, shift)

    # Set up LinearRTO with AffineModel
    y = cuqi.distribution.Gaussian(affine_model, 0.01**2)
    posterior_affine = cuqi.distribution.JointDistribution(x, y)(y=y_obs)

    # Set up LinearRTO with AffineModel
    sampler_affine = cuqi.sampler.LinearRTO(posterior_affine)

    # Sample with fixes seed
    np.random.seed(0)
    samples_affine = sampler_affine.sample(10, 2)

    # Check that the samples are the same
    assert np.allclose(samples_linear.samples, samples_affine.samples)

def test_RegularizedRTO_with_AffieModel_is_equivalent_to_LinearModel_and_shifted_data():
    """ Test that sampling with RegularizedRTO with an AffineModel is equivalent to sampling with LinearModel and shifted data. """

    # Define LinearModel and data
    A, y_obs, _ = cuqi.testproblem.Deconvolution1D().get_components()

    # Define Shift
    shift = np.random.rand(A.domain_dim)

    # Define Bayesian Problem
    x = cuqi.implicitprior.NonnegativeGMRF(np.zeros(A.domain_dim), 100)
    y = cuqi.distribution.Gaussian(A@x, 0.01**2)
    posterior = cuqi.distribution.JointDistribution(x, y)(y=y_obs-shift)

    # Set up RegularizedRTO with both models
    sampler_linear = cuqi.sampler.RegularizedLinearRTO(posterior)

    # Sample with fixes seed
    np.random.seed(0)
    samples_linear = sampler_linear.sample(10, 2)

    # Define AffineModel
    affine_model = cuqi.model.AffineModel(A, shift)

    # Set up RegularizedRTO with AffineModel
    y = cuqi.distribution.Gaussian(affine_model, 0.01**2)
    posterior_affine = cuqi.distribution.JointDistribution(x, y)(y=y_obs)

    # Set up RegularizedRTO with AffineModel
    sampler_affine = cuqi.sampler.RegularizedLinearRTO(posterior_affine)

    # Sample with fixes seed
    np.random.seed(0)
    samples_affine = sampler_affine.sample(10, 2)

    # Check that the samples are the same
    assert np.allclose(samples_linear.samples, samples_affine.samples)
        
def test_UGLA_with_AffineModel_is_equivalent_to_LinearModel_and_shifted_data():
    """ Test that sampling with UGLA with an AffineModel is equivalent to sampling with LinearModel and shifted data. """

    # Define LinearModel and data
    A, y_obs, _ = cuqi.testproblem.Deconvolution1D().get_components()

    # Define Shift
    shift = np.random.rand(A.domain_dim)

    # Define Bayesian Problem
    x = cuqi.distribution.LMRF(np.zeros(A.domain_dim), 0.01)
    y = cuqi.distribution.Gaussian(A@x, 0.01**2)
    posterior = cuqi.distribution.JointDistribution(x, y)(y=y_obs-shift)

    # Set up UGLA with both models
    sampler_linear = cuqi.sampler.UGLA(posterior)

    # Sample with fixes seed
    np.random.seed(0)
    samples_linear = sampler_linear.sample(10, 2)

    # Define AffineModel
    affine_model = cuqi.model.AffineModel(A, shift)

    # Set up UGLA with AffineModel
    y = cuqi.distribution.Gaussian(affine_model, 0.01**2)
    posterior_affine = cuqi.distribution.JointDistribution(x, y)(y=y_obs)

    # Set up UGLA with AffineModel
    sampler_affine = cuqi.sampler.UGLA(posterior_affine)

    # Sample with fixes seed
    np.random.seed(0)
    samples_affine = sampler_affine.sample(10, 2)

    # Check that the samples are the same
    assert np.allclose(samples_linear.samples, samples_affine.samples)



from cuqi.geometry import Geometry
import numpy as np

class _ProductGeometry(Geometry):
    """ A class for representing a product geometry. A product geometry
    represents the product space of multiple geometries of type :class:`Geometry`.
    See the example below for a product geometry of two geometries.

    Parameters
    ----------
    \*geometries : cuqi.geometry.Geometry
        The geometries to be combined into a product geometry. Each geometry
        is passed as a comma-separated argument.

    Example
    -------
    .. code-block:: python
        import numpy as np
        from cuqi.geometry import Continuous1D, Discrete
        from cuqi.experimental.geometry import _ProductGeometry
        geometry1 = Continuous1D(np.linspace(0, 1, 100))
        geometry2 = Discrete(["sound_speed"])
        product_geometry = _ProductGeometry(geometry1, geometry2)
    """

    def __init__(self, *geometries):
        self.geometries = geometries

    @property
    def geometries(self):
        """List of geometries that are combined to form the product geometry."""
        return self._geometries
    
    @geometries.setter
    def geometries(self, geometries):
        # Check if all geometries are of type Geometry.
        for g in geometries:
            if not isinstance(g, Geometry):
                raise TypeError(
                    "All geometries must be of type Geometry. "
                    "Received: {}".format(type(g))
                )
        self._geometries = geometries

    @property
    def fun_shape(self):
        """Shape of the function representation. Returns a tuple, where
        each element of the tuple is the shape of the function 
        representation of each geometry."""
        return tuple([g.fun_shape for g in self.geometries])
    
    @property
    def fun_dim(self):
        """Dimension of the function representation which is the sum of
        the function representation dimensions of each geometry."""
        return sum([g.fun_dim for g in self.geometries])

    @property
    def par_shape(self):
        """Shape of the parameter representation. Returns a tuple, where
        each element of the tuple is the shape of the parameter
        representation of each geometry."""
        return tuple([g.par_shape for g in self.geometries])
    
    @property
    def par_dim(self):
        """Dimension of the parameter representation which is the sum of
        the parameter representation dimensions of each geometry."""
        return sum(self.par_dim_list)
    
    @property
    def par_dim_list(self):
        """List of the parameter representation dimensions of each
        geometry. This property is useful for indexing a stacked parameter
        vector."""
        return [g.par_dim for g in self.geometries]
    
    @property
    def stacked_par_split_indices(self):
        """Indices at which the stacked parameter vector should be split
        to obtain the parameter vectors for each geometry. For example, if
        the stacked parameter vector is [1, 2, 3, 4, 5, 6] and the parameter
        vectors for each geometry are [1, 2], [3, 4], and [5, 6], then the
        split indices are [2, 4]"""
        return np.cumsum(self.par_dim_list[:-1])

    @property
    def number_of_geometries(self):
        """Number of geometries in the product geometry."""
        return len(self.geometries)
    
    def _split_par(self, par):
        """Splits a stacked parameter vector into parameter vectors for each
        geometry."""
        return tuple(np.split(par, self.stacked_par_split_indices))
    
    def _plot(self, values, **kwargs):
        """Plotting function for the product geometry."""
        raise NotImplementedError(
            f"Plotting not implemented for {self.__class__.__name__}.")

    def par2fun(self, *pars):
        """Converts parameter vector(s) into function values for each 
        geometry. The parameter vector can be stacked (all parameters are
        in one vector) or unstacked (one parameter vector corresponds to
        each geometry). In all cases, the order of the parameter vectors
        should follow the order of the geometries in the product, i.e., the
        first parameter vector corresponds to the first geometry and so on."""

        # If one argument is passed, then it is assumed that the parameter
        # vector is stacked and split it.
        # No effect if the parameter vector is already split and corresponds
        # to one geometry.
        if len(pars) == 1:
            pars = self._split_par(pars[0])

        # Convert parameter vectors to function values for each geometry.
        funvals = []
        for i, g in enumerate(self.geometries):
            funval_i = g.par2fun(pars[i])
            funvals.append(funval_i)
        return tuple(funvals)

    def fun2par(self, *funvals, stacked=False):
        """Converts (multiple) function values into the corresponding
        parameter vectors. If the flag stacked is set to True, then the
        parameter vectors are stacked into one vector. Otherwise, the
        parameter vectors are returned as a tuple. The order of function
        values should follow the order of the geometries in the product,
        i.e., the first function value corresponds to the first geometry
        and so on."""

        pars = []
        for i, g in enumerate(self.geometries):
            par_i = g.fun2par(funvals[i])
            pars.append(par_i)

        # stack parameters:
        if stacked:
            # if single sample
            if len(pars[0].shape) == 1:
                stacked_val = np.hstack(pars)
            elif len(pars[0].shape) == 2:
                stacked_val = np.vstack(pars)
            else:
                raise ValueError(
                    "Cannot stack parameter vectors with more than 2 dimensions."
                    )

        return  stacked_val if stacked else tuple(pars)

    def vec2fun(self, *funvecs):
        """Maps function vector representation, if available, to function 
        values. The order of the function vectors should follow the order of
        the geometries in the product, i.e., the first function vector
        corresponds to the first geometry and so on."""
        funvals = []  
        for i, g in enumerate(self.geometries):
            funvals.append(g.vec2fun(funvecs[i]))

        return tuple(funvals)
    
    def fun2vec(self, *funvals):
        """Maps function values to a vector representation of the function
        values, if available. The order of the function values should follow
        the order of the geometries in the product, i.e., the first function
        value corresponds to the first geometry and so on."""
        funvecs = []
        for i, g in enumerate(self.geometries):
            funvecs.append(g.fun2vec(funvals[i]))
        
        return tuple(funvecs)
        
    
    def __repr__(self, pad="") -> str:
        """Representation of the product geometry."""
        string = "{}(".format(self.__class__.__name__) + "\n"
        for g in self.geometries:
            string += pad + "    {}\n".format(g.__repr__())
        string += pad + ")"
        return string


import numpy as np
from cuqi.utilities import force_ndarray
from cuqi.distribution import Distribution

class SmoothedLaplace(Distribution):
    """ Smoothed Laplace distribution. 

    Defines a smoothed Laplace distribution given a location, a scale and a smoothing parameter
    beta. The smoothed Laplace distribution is defined as

    .. math::

        p(x) = \\frac{1}{2b} \exp\left(-\\frac{\sqrt{(x-\mu)^2 + \\beta}}{b}\\right),

    where :math:`\mu` is the location (mean), :math:`b` is the scale (decay) parameter and
    :math:`\\beta` is the smoothing parameter.

    The rate parameter is defined as :math:`\lambda = \\frac{1}{b}`.

    The variables of this Laplace distribution are independent identically distributed (i.i.d.).

    Parameters
    ----------
    location : scalar, list, tuple, or ndarray
        The location parameter of the distribution.

    scale : scalar, list, tuple, or ndarray
        The scale parameter of the distribution.
    
    beta : scalar
        The smoothing parameter of the distribution.

    """

    def __init__(self, location=None, scale=None, beta=1e-3, **kwargs):
        super().__init__(**kwargs)

        self.location = location
        self.scale = scale
        self.beta = beta

    @property
    def location(self):
        """ Location parameter """
        return self._location
    
    @location.setter
    def location(self, value):
        self._location = force_ndarray(value, flatten=True)

    @property
    def scale(self):
        """ Scale parameter """
        return self._scale
    
    @scale.setter
    def scale(self, value):
        self._scale = force_ndarray(value, flatten=True)

    @property
    def beta(self):
        """ Beta parameter """
        return self._beta
    
    @beta.setter
    def beta(self, value):
        self._beta = value

    def logpdf(self, x):
        """
        Computes the logarithm of the probability density function at the given values of x.
        """
        # x accepts scalar, list, tuple, or ndarray
        if isinstance(x, (float, int)):
            x = np.array([x])
        elif isinstance(x, (list, tuple)):
            x = np.array(x)
        return np.sum(np.log(0.5 / self.scale)) - np.sum(np.sqrt((x - self.location) ** 2 + self.beta) / self.scale)

    def gradient(self, x):
        """
        Computes the gradient of logpdf at the given values of x.
        """
        # x accepts scalar, list, tuple, or ndarray
        if isinstance(x, (float, int)):
            x = np.array([x])
        elif isinstance(x, (list, tuple)):
            x = np.array(x)
        return -np.array((x - self.location) / self.scale / np.sqrt((x - self.location) ** 2 + self.beta))

    def _sample(self, N=1, rng=None):
        """
        Generates random samples from the distribution.
        """
        raise NotImplementedError(f"sample is not implemented for {self.__class__.__name__}.")



from abc import ABC, abstractmethod
import os
import numpy as np
import pickle as pkl
import warnings
import cuqi
from cuqi.samples import Samples

try:
    from tqdm import tqdm
except ImportError:
    def tqdm(iterable, **kwargs):
        warnings.warn("Module mcmc: tqdm not found. Install tqdm to get sampling progress.")
        return iterable

class Sampler(ABC):
    """ Abstract base class for all samplers.
    
    Provides a common interface for all samplers. The interface includes methods for sampling, warmup and getting the samples in an object oriented way.

    Samples are stored in a list to allow for dynamic growth of the sample set. Returning samples is done by creating a new Samples object from the list of samples.

    The sampler maintains sets of state and history keys, which are used for features like checkpointing and resuming sampling.

    The state of the sampler represents all variables that are updated (replaced) in a Markov Monte Carlo step, e.g. the current point of the sampler.

    The history of the sampler represents all variables that are updated (appended) in a Markov Monte Carlo step, e.g. the samples and acceptance rates.

    Subclasses should ensure that any new variables that are updated in a Markov Monte Carlo step are added to the state or history keys.

    Saving and loading checkpoints saves and loads the state of the sampler (not the history).

    Batching samples via the batch_size parameter saves the sampler history to disk in batches of the specified size.

    Any other attribute stored as part of the sampler (e.g. target, initial_point) is not supposed to be updated
    during sampling and should not be part of the state or history.

    """

    _STATE_KEYS = {'current_point'}
    """ Set of keys for the state dictionary. """

    _HISTORY_KEYS = {'_samples', '_acc'}
    """ Set of keys for the history dictionary. """

    def __init__(self, target:cuqi.density.Density=None, initial_point=None, callback=None):
        """ Initializer for abstract base class for all samplers.

        Any subclassing samplers should simply store input parameters as part of the __init__ method. 

        The actual initialization of the sampler should be done in the _initialize method.
        
        Parameters
        ----------
        target : cuqi.density.Density
            The target density.

        initial_point : array-like, optional
            The initial point for the sampler. If not given, the sampler will choose an initial point.

        callback : callable, optional
            A function that will be called after each sampling step. It can be useful for monitoring the sampler during sampling.
            The function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature is: `callback(sampler, sample_index, num_of_samples)`.
        """

        self.target = target
        self.initial_point = initial_point
        self.callback = callback
        self._is_initialized = False

    def initialize(self):
        """ Initialize the sampler by setting and allocating the state and history before sampling starts. """

        if self._is_initialized:
            raise ValueError("Sampler is already initialized.")
        
        if self.target is None:
            raise ValueError("Cannot initialize sampler without a target density.")
        
        # Default values
        if self.initial_point is None:
            self.initial_point = self._get_default_initial_point(self.dim)

        # State variables
        self.current_point = self.initial_point

        # History variables
        self._samples = []
        self._acc = [ 1 ] # TODO. Check if we need to put 1 here.

        self._initialize() # Subclass specific initialization

        self._validate_initialization()

        self._is_initialized = True

    # ------------ Abstract methods to be implemented by subclasses ------------
    @abstractmethod
    def step(self):
        """ Perform one step of the sampler by transitioning the current point to a new point according to the sampler's transition kernel. """
        pass

    @abstractmethod
    def tune(self, skip_len, update_count):
        """ Tune the parameters of the sampler. This method is called after each step of the warmup phase.

        Parameters
        ----------
        skip_len : int
            Defines the number of steps in between tuning (i.e. the tuning interval).

        update_count : int
            The number of times tuning has been performed. Can be used for internal bookkeeping.
        
        """
        pass

    @abstractmethod
    def validate_target(self):
        """ Validate the target is compatible with the sampler. Called when the target is set. Should raise an error if the target is not compatible. """
        pass

    @abstractmethod
    def _initialize(self):
        """ Subclass specific sampler initialization. Called during the initialization of the sampler which is done before sampling starts. """
        pass

    # ------------ Public attributes ------------
    @property
    def dim(self) -> int:
        """ Dimension of the target density. """
        return self.target.dim

    @property
    def geometry(self) -> cuqi.geometry.Geometry:
        """ Geometry of the target density. """
        return self.target.geometry
    
    @property
    def target(self) -> cuqi.density.Density:
        """ Return the target density. """
        return self._target
    
    @target.setter
    def target(self, value):
        """ Set the target density. Runs validation of the target. """
        self._target = value
        if self._target is not None:
            self.validate_target()

    @property
    def current_point(self):
        """ The current point of the sampler. """
        return self._current_point

    @current_point.setter
    def current_point(self, value):
        """ Set the current point of the sampler. """
        self._current_point = value

    # ------------ Public methods ------------
    def get_samples(self) -> Samples:
        """ Return the samples. The internal data-structure for the samples is a dynamic list so this creates a copy. """
        return Samples(np.array(self._samples).T, self.target.geometry)
    
    def reinitialize(self):
        """ Re-initialize the sampler. This clears the state and history and initializes the sampler again by setting state and history to their original values. """

        # Loop over state and reset to None
        for key in self._STATE_KEYS:
            setattr(self, key, None)

        # Loop over history and reset to None
        for key in self._HISTORY_KEYS:
            setattr(self, key, None)

        self._is_initialized = False

        self.initialize()
    
    def save_checkpoint(self, path):
        """ Save the state of the sampler to a file. """

        self._ensure_initialized()

        state = self.get_state()

        # Convert all CUQIarrays to numpy arrays since CUQIarrays do not get pickled correctly
        for key, value in state['state'].items():
            if isinstance(value, cuqi.array.CUQIarray):
                state['state'][key] = value.to_numpy()

        with open(path, 'wb') as handle:
            pkl.dump(state, handle, protocol=pkl.HIGHEST_PROTOCOL)

    def load_checkpoint(self, path):
        """ Load the state of the sampler from a file. """

        self._ensure_initialized()

        with open(path, 'rb') as handle:
            state = pkl.load(handle)

        self.set_state(state)

    def sample(self, Ns, Nt=1, batch_size=0, sample_path='./CUQI_samples/') -> 'Sampler':
        """ Sample Ns samples from the target density.

        Parameters
        ----------
        Ns : int
            The number of samples to draw.
        
        Nt : int, optional, default=1
            The thinning interval. If Nt >= 1, every Nt'th sample is stored. The larger Nt, the fewer samples are stored.

        batch_size : int, optional
            The batch size for saving samples to disk. If 0, no batching is used. If positive, samples are saved to disk in batches of the specified size.

        sample_path : str, optional
            The path to save the samples. If not specified, the samples are saved to the current working directory under a folder called 'CUQI_samples'.

        """
        self._ensure_initialized()

        # Initialize batch handler
        if batch_size > 0:
            batch_handler = _BatchHandler(batch_size, sample_path)

        # Draw samples
        pbar = tqdm(range(Ns), "Sample: ")
        for idx in pbar:
            
            # Perform one step of the sampler
            acc = self.step()

            # Store samples
            self._acc.append(acc)
            if (Nt > 0) and ((idx + 1) % Nt == 0):
                self._samples.append(self.current_point)

            # display acc rate at progress bar
            pbar.set_postfix_str(f"acc rate: {np.mean(self._acc[-1-idx:]):.2%}")

            # Add sample to batch
            if batch_size > 0:
                batch_handler.add_sample(self.current_point)

            # Call callback function if specified            
            self._call_callback(idx, Ns)
                
        return self
    

    def warmup(self, Nb, Nt=1, tune_freq=0.1) -> 'Sampler':
        """ Warmup the sampler by drawing Nb samples.

        Parameters
        ----------
        Nb : int
            The number of samples to draw during warmup.

        Nt : int, optional, default=1
            The thinning interval. If Nt >= 1, every Nt'th sample is stored. The larger Nt, the fewer samples are stored.
            
        tune_freq : float, optional
            The frequency of tuning. Tuning is performed every tune_freq*Nb samples.

        """

        self._ensure_initialized()

        tune_interval = max(int(tune_freq * Nb), 1)

        # Draw warmup samples with tuning
        pbar = tqdm(range(Nb), "Warmup: ")
        for idx in pbar:

            # Perform one step of the sampler
            acc = self.step()

            # Tune the sampler at tuning intervals
            if (idx + 1) % tune_interval == 0:
                self.tune(tune_interval, idx // tune_interval) 

            # Store samples
            self._acc.append(acc)
            if (Nt > 0) and ((idx + 1) % Nt == 0):
                self._samples.append(self.current_point)

            # display acc rate at progress bar
            pbar.set_postfix_str(f"acc rate: {np.mean(self._acc[-1-idx:]):.2%}")

            # Call callback function if specified
            self._call_callback(idx, Nb)

        return self
    
    def get_state(self) -> dict:
        """ Return the state of the sampler. 

        The state is used when checkpointing the sampler.

        The state of the sampler is a dictionary with keys 'metadata' and 'state'.
        The 'metadata' key contains information about the sampler type.
        The 'state' key contains the state of the sampler.

        For example, the state of a "MH" sampler could be:

        state = {
            'metadata': {
                'sampler_type': 'MH'
            },
            'state': {
                'current_point': np.array([...]),
                'current_target_logd': -123.45,
                'scale': 1.0,
                ...
            }
        }
        """   
        state = {
            'metadata': {
                'sampler_type': self.__class__.__name__
            },
            'state': {
                key: getattr(self, key) for key in self._STATE_KEYS
            }
        }
        return state

    def set_state(self, state: dict):
        """ Set the state of the sampler.

        The state is used when loading the sampler from a checkpoint.

        The state of the sampler is a dictionary with keys 'metadata' and 'state'.

        For example, the state of a "MH" sampler could be:

        state = {
            'metadata': {
                'sampler_type': 'MH'
            },
            'state': {
                'current_point': np.array([...]),
                'current_target_logd': -123.45,
                'scale': 1.0,
                ...
            }
        }
        """
        if state['metadata']['sampler_type'] != self.__class__.__name__:
            raise ValueError(f"Sampler type in state dictionary ({state['metadata']['sampler_type']}) does not match the type of the sampler ({self.__class__.__name__}).")
        
        for key, value in state['state'].items():
            if key in self._STATE_KEYS:
                setattr(self, key, value)
            else:
                raise ValueError(f"Key {key} not recognized in state dictionary of sampler {self.__class__.__name__}.")
            
    def get_history(self) -> dict:
        """ Return the history of the sampler. """
        history = {
            'metadata': {
                'sampler_type': self.__class__.__name__
            },
            'history': {
                key: getattr(self, key) for key in self._HISTORY_KEYS
            }
        }
        return history
    
    def set_history(self, history: dict):
        """ Set the history of the sampler. """
        if history['metadata']['sampler_type'] != self.__class__.__name__:
            raise ValueError(f"Sampler type in history dictionary ({history['metadata']['sampler_type']}) does not match the type of the sampler ({self.__class__.__name__}).")
        
        for key, value in history['history'].items():
            if key in self._HISTORY_KEYS:
                setattr(self, key, value)
            else:
                raise ValueError(f"Key {key} not recognized in history dictionary of sampler {self.__class__.__name__}.")

    # ------------ Private methods ------------
    def _call_callback(self, sample_index, num_of_samples):
        """ Calls the callback function. Assumes input is sampler, sample index, and total number of samples """
        if self.callback is not None:
            self.callback(self, sample_index, num_of_samples)

    def _validate_initialization(self):
        """ Validate the initialization of the sampler by checking all state and history keys are set. """

        for key in self._STATE_KEYS:
            if getattr(self, key) is None:
                raise ValueError(f"Sampler state key {key} is not set after initialization.")

        for key in self._HISTORY_KEYS:
            if getattr(self, key) is None:
                raise ValueError(f"Sampler history key {key} is not set after initialization.")
            
    def _ensure_initialized(self):
        """ Ensure the sampler is initialized. If not initialize it. """
        if not self._is_initialized:
            self.initialize()
            
    def _get_default_initial_point(self, dim):
        """ Return the default initial point for the sampler. Defaults to an array of ones. """
        return np.ones(dim)
    
    def __repr__(self):
        """ Return a string representation of the sampler. """
        if self.target is None:
            return f"Sampler: {self.__class__.__name__} \n Target: None"
        else:
            msg = f"Sampler: {self.__class__.__name__} \n Target: \n \t {self.target} "
            
        if self._is_initialized:
            state = self.get_state()
            msg += f"\n Current state: \n"
            # Sort keys alphabetically
            keys = sorted(state['state'].keys())
            # Put _ in the end
            keys = [key for key in keys if key[0] != '_'] + [key for key in keys if key[0] == '_']
            for key in keys:
                value = state['state'][key]
                msg += f"\t {key}: {value} \n"
        return msg

class ProposalBasedSampler(Sampler, ABC):
    """ Abstract base class for samplers that use a proposal distribution. """

    _STATE_KEYS = Sampler._STATE_KEYS.union({'current_target_logd', 'scale'})

    def __init__(self, target=None, proposal=None, scale=1, **kwargs):
        """ Initializer for abstract base class for samplers that use a proposal distribution.

        Any subclassing samplers should simply store input parameters as part of the __init__ method.

        Initialization of the sampler should be done in the _initialize method.

        See :class:`Sampler` for additional details.

        Parameters
        ----------
        target : cuqi.density.Density
            The target density.

        proposal : cuqi.distribution.Distribution, optional
            The proposal distribution. If not specified, the default proposal is used.

        scale : float, optional
            The scale parameter for the proposal distribution.

        **kwargs : dict
            Additional keyword arguments passed to the :class:`Sampler` initializer.

        """

        super().__init__(target, **kwargs)
        self.proposal = proposal
        self.initial_scale = scale

    def initialize(self):
        """ Initialize the sampler by setting and allocating the state and history before sampling starts. """

        if self._is_initialized:
            raise ValueError("Sampler is already initialized.")
        
        if self.target is None:
            raise ValueError("Cannot initialize sampler without a target density.")
        
        # Default values
        if self.initial_point is None:
            self.initial_point = self._get_default_initial_point(self.dim)

        if self.proposal is None:
            self.proposal = self._default_proposal

        # State variables
        self.current_point = self.initial_point
        self.scale = self.initial_scale

        self.current_target_logd = self.target.logd(self.current_point)

        # History variables
        self._samples = []
        self._acc = [ 1 ] # TODO. Check if we need to put 1 here.

        self._initialize() # Subclass specific initialization

        self._validate_initialization()

        self._is_initialized = True

    @abstractmethod
    def validate_proposal(self):
        """ Validate the proposal distribution. """
        pass     

    @property
    def _default_proposal(self):
        """ Return the default proposal distribution. Defaults to a Gaussian distribution with zero mean and unit variance. """
        return cuqi.distribution.Gaussian(np.zeros(self.dim), 1)

    @property
    def proposal(self):
        """ The proposal distribution. """
        return self._proposal
    
    @proposal.setter
    def proposal(self, proposal):
        """ Set the proposal distribution. """
        self._proposal = proposal
        if self._proposal is not None:
            self.validate_proposal()


class _BatchHandler:
    """ Utility class to handle batching of samples. 
    
    If a batch size is specified, this class will save samples to disk in batches of the specified size.
     
    This is useful for very large sample sets that do not fit in memory.
     
    """
    
    def __init__(self, batch_size=0, sample_path='./CUQI_samples/'):

        if batch_size < 0:
            raise ValueError("Batch size should be a non-negative integer")

        self.sample_path = sample_path
        self._batch_size = batch_size
        self.current_batch = []
        self.num_batches_dumped = 0

    @property
    def sample_path(self):
        """ The path to save the samples. """
        return self._sample_path
    
    @sample_path.setter
    def sample_path(self, value):
        if not isinstance(value, str):
            raise TypeError("Sample path must be a string.")
        normalized_path = value.rstrip('/') + '/'
        if not os.path.isdir(normalized_path):
            try:
                os.makedirs(normalized_path, exist_ok=True)
            except Exception as e:
                raise ValueError(f"Could not create directory at {normalized_path}: {e}")
        self._sample_path = normalized_path

    def add_sample(self, sample):
        """ Add a sample to the batch if batching. If the batch is full, flush the batch to disk. """

        if self._batch_size <= 0:
            return  # Batching not used

        self.current_batch.append(sample)

        if len(self.current_batch) >= self._batch_size:
            self.flush()

    def flush(self):
        """ Flush the current batch of samples to disk. """

        if not self.current_batch:
            return  # No samples to flush

        # Save the current batch of samples
        batch_samples = np.array(self.current_batch)
        file_path = f'{self.sample_path}batch_{self.num_batches_dumped:04d}.npz'
        np.savez(file_path, samples=batch_samples, batch_id=self.num_batches_dumped)

        self.num_batches_dumped += 1
        self.current_batch = []  # Clear the batch after saving

    def finalize(self):
        """ Finalize the batch handler. Flush any remaining samples to disk. """
        self.flush()


from __future__ import annotations
import numpy as np
import time
from typing import Tuple

import cuqi
from cuqi import config
from cuqi.distribution import Distribution, Gaussian, InverseGamma, LMRF, GMRF, Lognormal, Posterior, Beta, JointDistribution, Gamma, ModifiedHalfNormal, CMRF
from cuqi.implicitprior import RegularizedGaussian, RegularizedGMRF
from cuqi.density import Density
from cuqi.model import LinearModel, Model
from cuqi.likelihood import Likelihood
from cuqi.geometry import _DefaultGeometry
from cuqi.utilities import ProblemInfo
from cuqi.array import CUQIarray
import warnings
import matplotlib.pyplot as plt

from copy import copy

class BayesianProblem(object):
    """ Representation of a Bayesian inverse problem defined by any number of densities (distributions and likelihoods), e.g.

    .. math::

        \\begin{align*}
        \mathrm{density}_1 &\sim \pi_1 \\newline
        \mathrm{density}_2 &\sim \pi_2 \\newline
                           &\\vdots
        \end{align*}

    The main goal of this class is to provide fully automatic methods for computing samples or point estimates of the Bayesian problem.

    This class uses :class:`~cuqi.distribution.JointDistribution` to model the Bayesian problem,
    and to condition on observed data. We term the resulting distribution the *target distribution*.

    Parameters
    ----------
    \*densities: Density
        The densities that represent the Bayesian Problem.
        Each density is passed as comma-separated arguments.
        Can be Distribution, Likelihood etc.

    \**data: ndarray, Optional
        Any potential observed data. The data should be passed
        as keyword arguments, where the keyword is the name of the
        density that the data is associated with.
        Data can alternatively be set using the :meth:`set_data` method,
        after the problem has been initialized.

    Examples
    --------

    **Basic syntax**

    Given distributions for ``x``, ``y`` and ``z``, we can define a Bayesian problem
    and set the observed data ``y=y_data`` as follows:

    .. code-block:: python

        BP = BayesianProblem(x, y, z).set_data(y=y_data)

    **Complete example**

    Consider a Bayesian inverse problem with a Gaussian prior and a Gaussian likelihood.
    Assume that the forward model is a linear model with a known matrix
    :math:`\mathbf{A}: \mathbf{x} \mapsto \mathbf{y}` and  that we have observed data
    :math:`\mathbf{y}=\mathbf{y}^\mathrm{obs}`. The Bayesian model can be summarized as

    .. math::

        \\begin{align*}
        \mathbf{x} &\sim \mathcal{N}(\mathbf{0}, 0.1^2 \mathbf{I}) \\newline
        \mathbf{y} &\sim \mathcal{N}(\mathbf{A}\mathbf{x}, 0.05^2 \mathbf{I}).
        \end{align*}
    
    Using the :class:`BayesianProblem` class, we can define the problem, set the data and
    compute samples from the posterior distribution associated with the problem as well as
    estimates such as the Maximum Likelihood (ML) and Maximum A Posteriori (MAP).

    .. note::

        In this case, we use a forward model from the :mod:`~cuqi.testproblem` module, but any
        custom forward model can added via the :mod:`~cuqi.model` module.

    .. code-block:: python

        # Import modules
        import cuqi
        import numpy as np
        import matplotlib.pyplot as plt

        # Deterministic forward model and data (1D convolution)
        A, y_data, probInfo = cuqi.testproblem.Deconvolution1D().get_components()

        # Bayesian model
        x = cuqi.distribution.Gaussian(np.zeros(A.domain_dim), 0.1)
        y = cuqi.distribution.Gaussian(A@x, 0.05)

        # Define Bayesian problem and set data
        BP = cuqi.problem.BayesianProblem(y, x).set_data(y=y_data)

        # Compute MAP estimate
        x_MAP = BP.MAP()

        # Compute samples from posterior
        x_samples = BP.sample_posterior(1000)

        # Plot results
        x_samples.plot_ci(exact=probInfo.exactSolution)
        plt.show()
        
        # Plot difference between MAP and sample mean
        (x_MAP - x_samples.mean()).plot()
        plt.title("MAP estimate - sample mean")
        plt.show()

    Notes
    -----

    In the simplest form the Bayesian problem represents a posterior distribution defined by two densities, i.e.,
    
    .. math::

        \pi_\mathrm{posterior}(\\boldsymbol{\\theta} \mid \mathbf{y}) \propto \pi_1(\mathbf{y} \mid \\boldsymbol{\\theta}) \pi_2(\\boldsymbol{\\theta}),

    where :math:`\pi_1(\mathbf{y} \mid \\boldsymbol{\\theta})` is a :class:`~cuqi.likelihood.Likelihood` function and :math:`\pi_2(\\boldsymbol{\\theta})` is a :class:`~cuqi.distribution.Distribution`.
    In this two-density case, the joint distribution reduces to a :class:`~cuqi.distribution.Posterior` distribution.

    Most functionality is currently only implemented for this simple case.

    """

    def get_components(self) -> Tuple[Model, CUQIarray, ProblemInfo]:
        """
        Method that returns the model, the data and additional information to be used in formulating the Bayesian problem.
        
        """

        problem_info = ProblemInfo() #Instead of a dict, we use our ProblemInfo dataclass.

        for key, value in vars(problem_info).items():
            if hasattr(self, key):
                setattr(problem_info,key,vars(self)[key])

        return self.model, self.data, problem_info

    def __init__(self, *densities: Density, **data: np.ndarray):
        self._target = JointDistribution(*densities)(**data)

    def set_data(self, **kwargs) -> BayesianProblem:
        """ Set the data of the problem. This conditions the underlying joint distribution on the data. """
        if not isinstance(self._target, JointDistribution):
            raise ValueError("Unable to set data for this problem. Maybe data is already set?")
        self._target = self._target(**kwargs)
        return self

    @property
    def data(self):
        """Extract the observed data from likelihood"""
        return self.likelihood.data

    @property
    def likelihood(self) -> Likelihood:
        """The likelihood function."""
        if not isinstance(self._target, Posterior):
            raise ValueError(f"Unable to extract likelihood from this problem. Current target is: \n {self._target}")
        return self._target.likelihood

    @likelihood.setter
    def likelihood(self, likelihood):
        if not isinstance(self._target, Posterior):
            raise ValueError(f"Unable to set likelihood for this problem. Current target is: \n {self._target}")
        self._target.likelihood = likelihood

    @property
    def prior(self) -> Distribution:
        """The prior distribution"""
        if not isinstance(self._target, Posterior):
            raise ValueError(f"Unable to extract prior from this problem. Current target is: \n {self._target}")
        return self._target.prior

    @prior.setter
    def prior(self, prior):
        if not isinstance(self._target, Posterior):
            raise ValueError(f"Unable to set prior for this problem. Current target is: \n {self._target}")
        self._target.prior = prior

    @property
    def model(self) -> Model:
        """Extract the cuqi model from likelihood."""
        return self.likelihood.model

    @property
    def posterior(self) -> Posterior:
        """Create posterior distribution from likelihood and prior."""
        if not isinstance(self._target, Posterior):
            raise ValueError(f"Unable to extract posterior for this problem. Current target is: \n {self._target}")
        return self._target

    def ML(self, disp=True, x0=None) -> CUQIarray:
        """ Compute the Maximum Likelihood (ML) estimate of the posterior.
        
        Parameters
        ----------

        disp : bool
            display info messages? (True or False).

        x0 : CUQIarray or ndarray
            User-specified initial guess for the solver. Defaults to a ones vector.

        Returns
        -------
        CUQIarray
            ML estimate of the posterior. Solver info is stored in the returned CUQIarray `info` attribute.
        
        """
        if disp:
            # Print warning to user about the automatic solver selection
            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("!!! Automatic solver selection is a work-in-progress !!!")
            print("!!!      Always validate the computed results.       !!!")
            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("")

        x_ML, solver_info = self._solve_max_point(self.likelihood, disp=disp, x0=x0)

        # Wrap the result in a CUQIarray and add solver info
        x_ML = cuqi.array.CUQIarray(x_ML, geometry=self.likelihood.geometry)
        x_ML.info = solver_info

        return x_ML


    def MAP(self, disp=True, x0=None) -> CUQIarray:
        """ Compute the Maximum A Posteriori (MAP) estimate of the posterior.
        
        Parameters
        ----------

        disp : bool
            display info messages? (True or False).

        x0 : CUQIarray or ndarray
            User-specified initial guess for the solver. Defaults to a ones vector.

        Returns
        -------
        CUQIarray
            MAP estimate of the posterior. Solver info is stored in the returned CUQIarray `info` attribute.
        
        """

        if disp:
            # Print warning to user about the automatic solver selection
            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("!!! Automatic solver selection is a work-in-progress !!!")
            print("!!!      Always validate the computed results.       !!!")
            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("")

        if self._check_posterior(self, Gaussian, Gaussian, LinearModel, max_dim=config.MAX_DIM_INV):
            if disp: print(f"Using direct MAP of Gaussian posterior. Only works for small-scale problems with dim<={config.MAX_DIM_INV}.")
            b  = self.data
            A  = self.model.get_matrix()
            Ce = self.likelihood.distribution.cov
            x0 = self.prior.mean
            Cx = self.prior.cov

            # If Ce and Cx are scalar, make them into matrices
            if np.size(Ce)==1:
                Ce = Ce.ravel()[0]*np.eye(self.model.range_dim)
            if np.size(Cx)==1:
                Cx = Cx.ravel()[0]*np.eye(self.model.domain_dim)

            #Basic MAP estimate using closed-form expression Tarantola 2005 (3.37-3.38)
            rhs = b-A@x0
            sysm = A@Cx@A.T+Ce
            x_MAP = x0 + Cx@(A.T@np.linalg.solve(sysm,rhs))
            solver_info = {"solver": "direct"}

        else: # If no specific implementation exists, use numerical optimization.
            x_MAP, solver_info = self._solve_max_point(self.posterior, disp=disp, x0=x0)

        # Wrap the result in a CUQIarray and add solver info
        x_MAP = cuqi.array.CUQIarray(x_MAP, geometry=self.posterior.geometry)
        x_MAP.info = solver_info
        return x_MAP

    def sample_posterior(self, Ns, Nb=None, callback=None, experimental=False) -> cuqi.samples.Samples:
        """Sample the posterior. Sampler choice and tuning is handled automatically.
        
        Parameters
        ----------
        Ns : int
            Number of samples to draw.

        Nb : int or None, *Optional*
            Number of burn-in samples. If not provided, 20% of the samples will be used for burn-in.

        callback : callable, *Optional*
            If set this function will be called after every sample.
            The signature of the callback function is `callback(sample, sample_index)`,
            where `sample` is the current sample and `sample_index` is the index of the sample.
            An example is shown in demos/demo31_callback.py.
            Note: if the parameter `experimental` is set to True, the callback function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature in the case is: `callback(sampler, sample_index, num_of_samples)`.

        experimental : bool, *Optional*
            If set to True, the sampler selection will use the samplers from the :mod:`cuqi.experimental.mcmc` module.

        Returns
        -------
        samples : cuqi.samples.Samples
            Samples from the posterior.
        
        """

        # Print warning to user about the automatic sampler selection
        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        print("!!! Automatic sampler selection is a work-in-progress. !!!")
        print("!!!       Always validate the computed results.        !!!")
        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        print("")

        if experimental:
            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("!!!  Using samplers from cuqi.experimental.mcmc  !!!")
            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("")

        # Set up burn-in if not provided
        if Nb is None:
            Nb = int(0.2*Ns)

        # If target is a joint distribution, try Gibbs sampling
        # This is still very experimental!
        if isinstance(self._target, JointDistribution):       
            return self._sampleGibbs(Ns, Nb, callback=callback, experimental=experimental)

        # For Gaussian small-scale we can use direct sampling
        if self._check_posterior(self, Gaussian, Gaussian, LinearModel, config.MAX_DIM_INV) and not self._check_posterior(self, GMRF):
            return self._sampleMapCholesky(Ns, callback)

        # For larger-scale Gaussian we use Linear RTO. TODO: Improve checking once we have a common Gaussian class.
        elif hasattr(self.prior,"sqrtprecTimesMean") and hasattr(self.likelihood.distribution,"sqrtprec") and isinstance(self.model,LinearModel):
            return self._sampleLinearRTO(Ns, Nb, callback, experimental=experimental)

        # For LMRF we use our awesome unadjusted Laplace approximation!
        elif self._check_posterior(self, LMRF, Gaussian):
            return self._sampleUGLA(Ns, Nb, callback, experimental=experimental)

        # If we have gradients, use NUTS!
        # TODO: Fix cases where we have gradients but NUTS fails (see checks)
        elif self._check_posterior(self, must_have_gradient=True) and not self._check_posterior(self, (Beta, InverseGamma, Lognormal)):
            return self._sampleNUTS(Ns, Nb, callback, experimental=experimental)

        # For Gaussians with non-linear model we use pCN
        elif self._check_posterior(self, (Gaussian, GMRF), Gaussian):
            return self._samplepCN(Ns, Nb, callback, experimental=experimental)
        
        # For Regularized Gaussians with linear models we use RegularizedLinearRTO
        elif self._check_posterior(self, (RegularizedGaussian, RegularizedGMRF), Gaussian, LinearModel):
            return self._sampleRegularizedLinearRTO(Ns, Nb, callback, experimental=experimental)

        else:
            raise NotImplementedError(f"Automatic sampler choice is not implemented for model: {type(self.model)}, likelihood: {type(self.likelihood.distribution)} and prior: {type(self.prior)} and dim {self.prior.dim}. Manual sampler choice can be done via the 'sampler' module. Posterior distribution can be extracted via '.posterior' of any testproblem (BayesianProblem).")

    def sample_prior(self, Ns, callback=None) -> cuqi.samples.Samples:
        """ Sample the prior distribution. Sampler choice and tuning is handled automatically. """

        # Try sampling prior directly
        try:
            return self.prior.sample(Ns)
        except NotImplementedError:
            pass

        # If no direct method exists redefine posterior to one with a constant likelihood and sample from posterior
        print("Using MCMC methods from sampler module to sample prior.")
        print("Make sure enough samples are drawn for convergence.")
        print("")

        # Create a copy of self
        prior_problem = copy(self)

        # Set likelihood to constant
        model = cuqi.model.LinearModel(lambda x: 0*x, lambda y: 0*y, self.model.range_geometry, self.model.domain_geometry)
        likelihood = cuqi.distribution.Gaussian(model, 1).to_likelihood(np.zeros(self.model.range_dim)) # p(y|x)=constant
        prior_problem.likelihood = likelihood
        
        # Set up burn-in
        Nb = int(0.2*Ns)

        # Now sample prior problem
        return prior_problem.sample_posterior(Ns, Nb, callback)

    def UQ(self, Ns=1000, Nb=None, percent=95, exact=None, experimental=False) -> cuqi.samples.Samples:
        """ Run an Uncertainty Quantification (UQ) analysis on the Bayesian problem and provide a summary of the results.
        
        Parameters
        ----------
        Ns : int, *Optional*
            Number of samples to draw. Defaults to 1000.

        Nb : int, *Optional*
            Number of burn-in samples. If not provided, 20% of the samples will be used for burn-in.
        
        exact : ndarray or dict[str, ndarray], *Optional*
            Exact solution to the problem. If provided the summary will include a comparison to the exact solution.
            If a dict is provided, the keys should be the names of the variables and the values should be the exact solution for each variable.
        
        percent : float, *Optional*
            The credible interval to plot. Defaults to 95%.

        experimental : bool, *Optional*
            If set to True, the sampler selection will use the samplers from the :mod:`cuqi.experimental.mcmc` module.

        Returns
        -------
        samples : cuqi.samples.Samples
            Samples from the posterior. The samples can be used to compute further statistics and plots.
        """
        print(f"Computing {Ns} samples")
        samples = self.sample_posterior(Ns, Nb, experimental=experimental)

        print("Plotting results")
        # Gibbs case
        if isinstance(samples, dict):
            for key, value in samples.items():
                if exact is not None and key in exact:
                    self._plot_UQ_for_variable(
                        value, percent=percent, exact=exact[key])
                else:
                    self._plot_UQ_for_variable(
                        value, percent=percent, exact=None)
        # Single parameter case
        else:
            self._plot_UQ_for_variable(
                samples, percent=percent, exact=exact)

        return samples

    def _plot_UQ_for_variable(
            self, samples: cuqi.samples.Samples, percent=None, exact=None):
        """ Do a fitting UQ plot for a single variable given by samples. """
        # Potentially extract exact solution
        if exact is None and hasattr(self, 'exactSolution'):
            exact = self.exactSolution

        # Plot traces for single parameters
        if samples.shape[0] == 1:
            samples.plot_trace(exact=exact)
        else: # Else plot credible intervals
            # If plot_ci throws a NotImplementedError (likely coming from 
            # _plot_envelope method), we try to plot the CI for the parameters
            # instead and plot the mean and the variance using the function
            # representation of the samples geometry.
            try:
                samples.plot_ci(percent=percent, exact=exact)
            except NotImplementedError as nie:
                print(
                    "Unable to plot CI for samples with the underlying " +
                    f"geometry: {samples.geometry}. Plotting the CI for the " +
                    "parameters instead.")
                self._alternative_plot_UQ_for_variable(
                    samples, percent=percent, exact=exact)

    def _alternative_plot_UQ_for_variable(
            self, samples: cuqi.samples.Samples, percent=None, exact=None):
        """ Alternative visualization for UQ analysis used when plot_ci
        method fails for the given samples geometry. """
        samples.plot_ci(percent=percent, exact=exact, plot_par=True)

        plt.figure()
        samples.plot_mean()
        plt.title("Sample parameter mean converted\nto function representation")

        plt.figure()
        samples.funvals.vector.plot_mean()
        plt.title("Sample mean of function representation")

        plt.figure()
        samples.plot_variance()
        plt.title(
            "Sample parameter variance converted\nto function representation")
        
        plt.figure()
        samples.funvals.vector.plot_variance()
        plt.title("Sample variance of function representation")

    def _sampleLinearRTO(self, Ns, Nb, callback=None, experimental=False):

        if experimental:

            print("Using cuqi.experimental.mcmc LinearRTO sampler.")
            print(f"burn-in: {Nb/Ns*100:g}%")

            sampler = cuqi.experimental.mcmc.LinearRTO(self.posterior, callback=callback)

            ti = time.time()

            sampler.warmup(Nb)
            sampler.sample(Ns)
            samples = sampler.get_samples().burnthin(Nb)

            print('Elapsed time:', time.time() - ti)

        else:

            print("Using cuqi.sampler LinearRTO sampler.")
            print(f"burn-in: {Nb/Ns*100:g}%")

            # Start timing
            ti = time.time()

            # Sample
            sampler = cuqi.sampler.LinearRTO(self.posterior, callback=callback)
            samples = sampler.sample(Ns, Nb)

            # Print timing
            print('Elapsed time:', time.time() - ti)

        return samples

    def _sampleMapCholesky(self, Ns, callback=None):
        print(f"Using direct sampling of Gaussian posterior. Only works for small-scale problems with dim<={config.MAX_DIM_INV}.")
        print("No burn-in needed for direct sampling.")

        # Start timing
        ti = time.time()

        b  = self.data
        A  = self.model.get_matrix()
        Ce = self.likelihood.distribution.cov
        x0 = self.prior.mean
        Cx = self.prior.cov

        # If Ce and Cx are scalar, make them into matrices
        if np.size(Ce)==1:
            Ce = Ce.ravel()[0]*np.eye(self.model.range_dim)
        if np.size(Cx)==1:
            Cx = Cx.ravel()[0]*np.eye(self.model.domain_dim)

        # Preallocate samples
        n = self.prior.dim 
        x_s = np.zeros((n,Ns))

        x_map = self.MAP(disp=False) #Compute MAP estimate
        C = np.linalg.inv(A.T@(np.linalg.inv(Ce)@A)+np.linalg.inv(Cx))
        L = np.linalg.cholesky(C)
        for s in range(Ns):
            x_s[:,s] = x_map.parameters + L@np.random.randn(n)
            # display iterations 
            if (s % 5e2) == 0:
                print("\r",'Sample', s, '/', Ns, end="")
            
            # Callback function
            if callback is not None:
                callback(x_s[:,s], s)

        print("\r",'Sample', s+1, '/', Ns)
        print('Elapsed time:', time.time() - ti)
        
        return cuqi.samples.Samples(x_s,self.model.domain_geometry)
    
    def _sampleCWMH(self, Ns, Nb, callback=None, experimental=False):

        if experimental:

            print("Using cuqi.experimental.mcmc Component-wise Metropolis-Hastings (CWMH) sampler.")
            print(f"burn-in: {Nb/Ns*100:g}%, scale: 0.05, x0: 0.5 (vector)")

            scale = 0.05*np.ones(self.prior.dim)
            x0 = 0.5*np.ones(self.prior.dim)

            sampler = cuqi.experimental.mcmc.CWMH(self.posterior, scale, x0, callback=callback)

            ti = time.time()

            sampler.warmup(Nb)
            sampler.sample(Ns)
            x_s = sampler.get_samples().burnthin(Nb)

            print('Elapsed time:', time.time() - ti)

        else:

            print("Using cuqi.sampler Component-wise Metropolis-Hastings (CWMH) sampler (sample_adapt)")
            print(f"burn-in: {Nb/Ns*100:g}%, scale: 0.05, x0: 0.5 (vector)")

            # Dimension
            n = self.prior.dim
            
            # Set up target and proposal
            def proposal(x_t, sigma): return np.random.normal(x_t, sigma)

            # Set up sampler
            scale = 0.05*np.ones(n)
            x0 = 0.5*np.ones(n)
            MCMC = cuqi.sampler.CWMH(self.posterior, proposal, scale, x0, callback=callback)
            
            # Run sampler
            ti = time.time()
            x_s = MCMC.sample_adapt(Ns,Nb); #ToDo: Make results class
            print('Elapsed time:', time.time() - ti)
        
        return x_s

    def _samplepCN(self, Ns, Nb, callback=None, experimental=False):

        if experimental:

            print("Using cuqi.experimental.mcmc preconditioned Crank-Nicolson (pCN) sampler.")
            print(f"burn-in: {Nb/Ns*100:g}%, scale: 0.02")

            scale = 0.02

            sampler = cuqi.experimental.mcmc.PCN(self.posterior, scale, callback=callback)

            ti = time.time()

            sampler.warmup(Nb)
            sampler.sample(Ns)
            x_s = sampler.get_samples().burnthin(Nb)

            print('Elapsed time:', time.time() - ti)

        else:

            print("Using cuqi.sampler preconditioned Crank-Nicolson (pCN) sampler (sample_adapt)")
            print(f"burn-in: {Nb/Ns*100:g}%, scale: 0.02")

            scale = 0.02
            
            MCMC = cuqi.sampler.pCN(self.posterior, scale, callback=callback)      
            
            #Run sampler
            ti = time.time()
            x_s = MCMC.sample_adapt(Ns, Nb)
            print('Elapsed time:', time.time() - ti)
       
        return x_s

    def _sampleNUTS(self, Ns, Nb, callback=None, experimental=False):

        if experimental:

            print("Using cuqi.experimental.mcmc No-U-Turn (NUTS) sampler.")
            print(f"burn-in: {Nb/Ns*100:g}%")

            sampler = cuqi.experimental.mcmc.NUTS(self.posterior, callback=callback)

            ti = time.time()

            sampler.warmup(Nb)
            sampler.sample(Ns)
            x_s = sampler.get_samples().burnthin(Nb)

            print('Elapsed time:', time.time() - ti)

        else:

            print("Using cuqi.sampler No-U-Turn (NUTS) sampler")
            print(f"burn-in: {Nb/Ns*100:g}%")
            
            MCMC = cuqi.sampler.NUTS(self.posterior, callback=callback)
            
            # Run sampler
            ti = time.time()
            x_s = MCMC.sample_adapt(Ns,Nb)
            print('Elapsed time:', time.time() - ti)
        
        return x_s

    def _sampleUGLA(self, Ns, Nb, callback=None, experimental=False):

        if experimental:

            print("Using cuqi.experimental.mcmc Unadjusted Gaussian Laplace Approximation (UGLA) sampler.")
            print(f"burn-in: {Nb/Ns*100:g}%")

            sampler = cuqi.experimental.mcmc.UGLA(self.posterior, callback=callback)

            ti = time.time()

            sampler.warmup(Nb)
            sampler.sample(Ns)
            samples = sampler.get_samples().burnthin(Nb)

            print('Elapsed time:', time.time() - ti)

        else:

            print("Using cuqi.sampler UGLA sampler")
            print(f"burn-in: {Nb/Ns*100:g}%")

            # Start timing
            ti = time.time()

            # Sample
            sampler = cuqi.sampler.UGLA(self.posterior, callback=callback)
            samples = sampler.sample(Ns, Nb)

            # Print timing
            print('Elapsed time:', time.time() - ti)

        return samples
    
    def _sampleRegularizedLinearRTO(self, Ns, Nb, callback=None, experimental=False):

        if experimental:

            print("Using cuqi.experimental.mcmc Regularized LinearRTO sampler.")
            print(f"burn-in: {Nb/Ns*100:g}%")

            sampler = cuqi.experimental.mcmc.RegularizedLinearRTO(self.posterior, maxit=100, stepsize = "automatic", abstol=1e-10, callback=callback)

            ti = time.time()

            sampler.warmup(Nb)
            sampler.sample(Ns)
            samples = sampler.get_samples().burnthin(Nb)

            print('Elapsed time:', time.time() - ti)
        
        else:

            print("Using cuqi.sampler Regularized LinearRTO sampler.")
            print(f"burn-in: {Nb/Ns*100:g}%")

            # Start timing
            ti = time.time()

            # Sample
            sampler = cuqi.sampler.RegularizedLinearRTO(self.posterior, maxit=100, stepsize = "automatic", abstol=1e-10, callback=callback)
            samples = sampler.sample(Ns, Nb)

            # Print timing
            print('Elapsed time:', time.time() - ti)

        return samples

    def _solve_max_point(self, density, disp=True, x0=None):
        """ This is a helper function for point estimation of the maximum of a density (e.g. posterior or likelihood) using solver module.
        
        Parameters
        ----------
        density : Density (Distribution or Likelihood)
            The density or likelihood to compute the maximum point of the negative log.

        disp : bool
            display info messages? (True or False).
        """
        
        # Get the function to minimize (negative log-likelihood or negative log-posterior)
        def func(x): return -density.logd(x)

        # Initial value if not given
        if x0 is None:
            x0 = np.ones(self.model.domain_dim)

        # Get the gradient (if available)
        try: 
            density.gradient(x0)
            def gradfunc(x): return -density.gradient(x)
            if disp: print("Optimizing with exact gradients")
        except (NotImplementedError, AttributeError):
            gradfunc = None
            if disp: print("Optimizing with approximate gradients.") 

        # Compute point estimate
        if self._check_posterior(self, CMRF, must_have_gradient=True): # Use L-BFGS-B for CMRF prior as it has better performance for this multi-modal posterior
            if disp: print(f"Using scipy.optimize.L_BFGS_B on negative log of {density.__class__.__name__}")
            if disp: print("x0: ones vector")
            solver = cuqi.solver.ScipyLBFGSB(func, x0, gradfunc=gradfunc)
        else:
            if disp: print(f"Using scipy.optimize.minimize on negative log of {density.__class__.__name__}")
            if disp: print("x0: ones vector")
            solver = cuqi.solver.ScipyMinimizer(func, x0, gradfunc=gradfunc)

        x_MAP, solver_info = solver.solve()

        # Add info on solver choice
        solver_info["solver"] = "L-BFGS-B"

        return x_MAP, solver_info

    def _check_geometries_consistency(self, geom1, geom2, fail_msg):
        """checks geom1 and geom2 consistency . If both are of type `_DefaultGeometry` they need to be equal. If one of them is of `_DefaultGeometry` type, it will take the value of the other one. If both of them are user defined, they need to be consistent"""
        if isinstance(geom1,_DefaultGeometry):
            if isinstance(geom2,_DefaultGeometry):
                if geom1 == geom2:
                    return geom1,geom2
            else: 
                return geom2, geom2
        else:
            if isinstance(geom2,_DefaultGeometry):
                return geom1,geom1
            else:
                if geom1 == geom2:
                    return geom1,geom2
        raise Exception(fail_msg)

    @staticmethod
    def _check_posterior(posterior, prior_type=None, likelihood_type=None, model_type=None, max_dim=None, must_have_gradient=False):
        """Returns true if components of the posterior reflects the types (can be tuple of types) given as input."""
        # Prior check
        if prior_type is None:
            P = True
        else:
            P = isinstance(posterior.prior, prior_type)

        # Likelihood check
        if likelihood_type is None:
            L = True
        else:
            L = isinstance(posterior.likelihood.distribution, likelihood_type)

        # Model check
        if model_type is None:
            M = True
        else:
            M = isinstance(posterior.model, model_type)

        #Dimension check
        if max_dim is None:
            D = True
        else:
            D = posterior.model.domain_dim<=max_dim and posterior.model.range_dim<=max_dim

        # Require gradient?
        if must_have_gradient:
            try: 
                posterior.posterior.gradient(np.zeros(posterior.posterior.dim))
                G = True
            except (NotImplementedError, AttributeError):
                G = False
        else:
            G = True

        return L and P and M and D and G

    def _sampleGibbs(self, Ns, Nb, callback=None, experimental=False):
        """ This is a helper function for sampling from the posterior using Gibbs sampler. """

        if experimental:

            print("Using cuqi.experimental.mcmc HybridGibbs sampler")
            print(f"burn-in: {Nb/Ns*100:g}%")
            print("")

            # Start timing
            ti = time.time()

            # Sampling strategy
            sampling_strategy = self._determine_sampling_strategy(experimental=True)

            sampler = cuqi.experimental.mcmc.HybridGibbs(
                self._target, sampling_strategy, callback=callback)
            sampler.warmup(Nb)
            sampler.sample(Ns)
            samples = sampler.get_samples()
            # Dict with Samples objects for each parameter
            # Now apply burnthin to each value in dict
            for key, value in samples.items():
                samples[key] = value.burnthin(Nb)
            
            # Print timing
            print('Elapsed time:', time.time() - ti)

        else:

            print("Using Gibbs sampler")
            print(f"burn-in: {Nb/Ns*100:g}%")
            print("")

            if callback is not None:
                raise NotImplementedError("Callback not implemented for Gibbs sampler. It is only implemented for experimental Gibbs sampler.")

            # Start timing
            ti = time.time()

            # Sampling strategy
            sampling_strategy = self._determine_sampling_strategy()

            sampler = cuqi.sampler.Gibbs(self._target, sampling_strategy)
            samples = sampler.sample(Ns, Nb)

            # Print timing
            print('Elapsed time:', time.time() - ti)

        return samples


    def _determine_sampling_strategy(self, experimental=False):
        """ This is a helper function for determining the sampling strategy for Gibbs sampler.
        
        It is still very experimental and not very robust.
        
        """

        # We determine sampling strategy by sequentially conditioning each variable on the others.
        # We then re-use the _check_posterior method to select the best sampler for each variable.
        # In the future we may consider refactoring these methods into one more robust way of
        # determining the sampling strategy.

        # Joint distribution and parameters
        joint = self._target
        par_names = joint.get_parameter_names()

        # Go through each parameter and condition on the others, then select the best sampler
        sampling_strategy = {}
        for par_name in par_names:

            # Dict of all other parameters to condition on with ones vector as initial value
            other_params = {par_name_: np.ones(joint.get_density(par_name_).dim) for par_name_ in par_names if par_name_ != par_name}

            # Condition on all other parameters to get target conditional distribution
            cond_target = joint(**other_params)

            # If not Posterior, we cant get sampling strategy (for now)
            if not isinstance(cond_target, Posterior):
                raise NotImplementedError(f"Unable to determine sampling strategy for {par_name} with target {cond_target}")

            # Gamma or ModifiedHalfNormal prior, Gaussian or RegularizedGaussian likelihood -> Conjugate
            if self._check_posterior(cond_target, (Gamma, ModifiedHalfNormal), (Gaussian, GMRF, RegularizedGaussian, RegularizedGMRF)):
                if experimental:
                    sampling_strategy[par_name] = cuqi.experimental.mcmc.Conjugate()
                else:
                    sampling_strategy[par_name] = cuqi.sampler.Conjugate

            # Gamma prior, LMRF likelihood -> ConjugateApprox
            elif self._check_posterior(cond_target, Gamma, LMRF):
                if experimental:
                    sampling_strategy[par_name] = cuqi.experimental.mcmc.ConjugateApprox()
                else:
                    sampling_strategy[par_name] = cuqi.sampler.ConjugateApprox

            # Gaussian prior, Gaussian likelihood, Linear model -> LinearRTO
            elif self._check_posterior(cond_target, (Gaussian, GMRF), Gaussian, LinearModel):
                if experimental:
                    sampling_strategy[par_name] = cuqi.experimental.mcmc.LinearRTO()
                else:
                    sampling_strategy[par_name] = cuqi.sampler.LinearRTO

            # Implicit Regularized Gaussian prior, Gaussian likelihood, linear model -> RegularizedLinearRTO
            elif self._check_posterior(cond_target, (RegularizedGaussian, RegularizedGMRF), Gaussian, LinearModel):
                if experimental:
                    sampling_strategy[par_name] = cuqi.experimental.mcmc.RegularizedLinearRTO()
                else:
                    sampling_strategy[par_name] = cuqi.sampler.RegularizedLinearRTO

            # LMRF prior, Gaussian likelihood, Linear model -> UGLA
            elif self._check_posterior(cond_target, LMRF, Gaussian, LinearModel):
                if experimental:
                    sampling_strategy[par_name] = cuqi.experimental.mcmc.UGLA()
                else:
                    sampling_strategy[par_name] = cuqi.sampler.UGLA

            else:
                raise NotImplementedError(f"Unable to determine sampling strategy for {par_name} with target {cond_target}")

        print("Automatically determined sampling strategy:")
        for dist_name, strategy in sampling_strategy.items():
            if experimental:
                print(f"\t{dist_name}: {strategy.__class__.__name__} (mcmc.experimental)")
            else:
                print(f"\t{dist_name}: {strategy.__name__}")
        print("")

        return sampling_strategy

    def __repr__(self):
        return f"BayesianProblem with target: \n {self._target}"



import numpy as np
import cuqi
from cuqi.experimental.mcmc import ProposalBasedSampler


class MH(ProposalBasedSampler):
    """ Metropolis-Hastings (MH) sampler.

    Parameters
    ----------
    target : cuqi.density.Density
        Target density or distribution.

    proposal : cuqi.distribution.Distribution or callable
        Proposal distribution. If None, a random walk MH is used (i.e., Gaussian proposal with identity covariance).

    scale : float
        Scaling parameter for the proposal distribution.

    kwargs : dict
        Additional keyword arguments to be passed to the base class :class:`ProposalBasedSampler`.

    """

    _STATE_KEYS = ProposalBasedSampler._STATE_KEYS.union({'scale', '_scale_temp'})

    def __init__(self, target=None, proposal=None, scale=1, **kwargs):
        super().__init__(target, proposal=proposal, scale=scale, **kwargs)

    def _initialize(self):
        # Due to a bug? in old MH, we must keep track of this extra variable to match behavior.
        self._scale_temp = self.scale

    def validate_target(self):
        # Fail only when there is no log density, which is currently assumed to be the case in case NaN is returned.
        if np.isnan(self.target.logd(self._get_default_initial_point(self.dim))):
            raise ValueError("Target does not have valid logd")

    def validate_proposal(self):
        if not isinstance(self.proposal, cuqi.distribution.Distribution):
            raise ValueError("Proposal must be a cuqi.distribution.Distribution object")
        if not self.proposal.is_symmetric:
            raise ValueError("Proposal must be symmetric")

    def step(self):
        # propose state
        xi = self.proposal.sample(1)   # sample from the proposal
        x_star = self.current_point + self.scale*xi.flatten()   # MH proposal

        # evaluate target
        target_eval_star = self.target.logd(x_star)

        # ratio and acceptance probability
        ratio = target_eval_star - self.current_target_logd # proposal is symmetric
        alpha = min(0, ratio)

        # accept/reject
        u_theta = np.log(np.random.rand())
        acc = 0
        if (u_theta <= alpha) and \
           (not np.isnan(target_eval_star)) and \
           (not np.isinf(target_eval_star)):
            self.current_point = x_star
            self.current_target_logd = target_eval_star
            acc = 1
        
        return acc

    def tune(self, skip_len, update_count):
        hat_acc = np.mean(self._acc[-skip_len:])

        # d. compute new scaling parameter
        zeta = 1/np.sqrt(update_count+1)   # ensures that the variation of lambda(i) vanishes

        # We use self._scale_temp here instead of self.scale in update. This might be a bug,
        # but is equivalent to old MH
        self._scale_temp = np.exp(np.log(self._scale_temp) + zeta*(hat_acc-0.234))

        # update parameters
        self.scale = min(self._scale_temp, 1)



from ._regularizedGaussian import RegularizedGaussian, ConstrainedGaussian, NonnegativeGaussian
from ._regularizedGMRF import RegularizedGMRF, ConstrainedGMRF, NonnegativeGMRF
from ._regularizedUnboundedUniform import RegularizedUnboundedUniform
from ._restorator import RestorationPrior, MoreauYoshidaPrior, TweediePrior




import numpy as np
from cuqi.experimental.mcmc import Conjugate
from cuqi.experimental.mcmc._conjugate import _ConjugatePair, _get_conjugate_parameter, _check_conjugate_parameter_is_scalar_reciprocal
from cuqi.distribution import LMRF, Gamma
import scipy as sp

class ConjugateApprox(Conjugate):
    """ Approximate Conjugate sampler

    Sampler for sampling a posterior distribution where the likelihood and prior can be approximated
    by a conjugate pair.

    Currently supported pairs are:
    - (LMRF, Gamma): Approximated by (Gaussian, Gamma) where Gamma is defined on the inverse of the scale parameter of the LMRF distribution.

    Gamma distribution must be univariate.

    LMRF likelihood must have zero mean.

    For more details on conjugacy see :class:`Conjugate`.

    """

    def _set_conjugatepair(self):
        """ Set the conjugate pair based on the likelihood and prior. This requires target to be set. """
        if isinstance(self.target.likelihood.distribution, LMRF) and isinstance(self.target.prior, Gamma):
            self._conjugatepair = _LMRFGammaPair(self.target)
        else:
            raise ValueError(f"Conjugacy is not defined for likelihood {type(self.target.likelihood.distribution)} and prior {type(self.target.prior)}, in CUQIpy")


class _LMRFGammaPair(_ConjugatePair):
    """ Implementation of the conjugate pair (LMRF, Gamma) """

    def validate_target(self):
        if not self.target.prior.dim == 1:
            raise ValueError("Approximate conjugate sampler only works with univariate Gamma prior")
        
        if np.sum(self.target.likelihood.distribution.location) != 0:
            raise ValueError("Approximate conjugate sampler only works with zero mean LMRF likelihood")
        
        key_value_pairs = _get_conjugate_parameter(self.target)
        if len(key_value_pairs) != 1:
            raise ValueError(f"Multiple references to conjugate parameter {self.target.prior.name} found in likelihood. Only one occurance is supported.")
        for key, value in key_value_pairs:
            if key == "scale":
                if not _check_conjugate_parameter_is_scalar_reciprocal(value):
                    raise ValueError("Approximate conjugate sampler only works with Gamma prior on the inverse of the scale parameter of the LMRF likelihood")
            else:
                raise ValueError(f"No approximate conjugacy defined for likelihood {type(self.target.likelihood.distribution)} and prior {type(self.target.prior)}, in CUQIpy")
        
    def conjugate_distribution(self):
        # Extract variables
        # Here we approximate the LMRF with a Gaussian

        # Extract diff_op from target likelihood
        D = self.target.likelihood.distribution._diff_op
        n = D.shape[0]

        # Gaussian approximation of LMRF prior as function of x_k
        # See Uribe et al. (2022) for details
        # Current has a zero mean assumption on likelihood! TODO
        beta=1e-5
        def Lk_fun(x_k):
            dd =  1/np.sqrt((D @ x_k)**2 + beta*np.ones(n))
            W = sp.sparse.diags(dd)
            return W.sqrt() @ D

        x = self.target.likelihood.data                 #x
        d = len(x)                                      #d
        Lx = Lk_fun(x)@x                                #Lx
        alpha = self.target.prior.shape                 #alpha
        beta = self.target.prior.rate                   #beta

        # Create Gamma distribution and sample
        return Gamma(shape=d+alpha, rate=np.linalg.norm(Lx)**2+beta)



import numpy as np
import scipy.stats as sps
from cuqi.geometry import _get_identity_geometries
from cuqi.utilities import force_ndarray
from cuqi.distribution import Distribution


class Beta(Distribution):
    """
    Multivariate beta distribution of independent random variables x_i. Each is distributed according to the PDF function

    .. math::

        f(x) = x^{(\\alpha-1)}(1-x)^{(\\beta-1)}\Gamma(\\alpha+\\beta) / (\Gamma(\\alpha)\Gamma(\\beta))

    where :math:`\Gamma` is the Gamma function.

    Parameters
    ------------
    alpha: float or array_like
           The shape parameter :math:`\\alpha` of the beta distribution.

    beta: float or array_like
          The shape parameter :math:`\\beta` of the beta distribution.
    
    Example
    -------
    .. code-block:: python

        # % Generate a beta distribution
        import numpy as np
        import cuqi
        import matplotlib.pyplot as plt
        alpha = 0.5
        beta  = 0.5
        rng = np.random.RandomState(1)
        x = cuqi.distribution.Beta(alpha, beta)
        samples = x.sample(1000, rng=rng)
        samples.hist_chain(0, bins=70)

    """
    def __init__(self, alpha=None, beta=None, is_symmetric=False, **kwargs):
        super().__init__(is_symmetric=is_symmetric, **kwargs)
        self.alpha = force_ndarray(alpha, flatten=True)
        self.beta = force_ndarray(beta, flatten=True)

    def logpdf(self, x):

        # Check bounds
        if np.any(x<=0) or np.any(x>=1) or np.any(self.alpha<=0) or np.any(self.beta<=0):
            return -np.inf

        # Compute logpdf
        return np.sum(sps.beta.logpdf(x, a=self.alpha, b=self.beta))

    def cdf(self, x):

        # Check bounds
        if np.any(x<=0) or np.any(x>=1) or np.any(self.alpha<=0) or np.any(self.beta<=0):
            return 0

        # Compute logpdf
        return np.prod(sps.beta.cdf(x, a=self.alpha, b=self.beta))

    def _sample(self, N=1, rng=None):
        return sps.beta.rvs(a=self.alpha, b=self.beta, size=(N,self.dim), random_state=rng).T

    def _gradient(self, x, *args, **kwargs):
        #Avoid complicated geometries that change the gradient.
        if not type(self.geometry) in _get_identity_geometries():
            raise NotImplementedError("Gradient not implemented for distribution {} with geometry {}".format(self,self.geometry))
        
        #Computing the gradient for conditional InverseGamma distribution is not supported yet    
        if self.is_cond:
            raise NotImplementedError(f"Gradient is not implemented for {self} with conditioning variables {self.get_conditioning_variables()}")
        
        # Check bounds (return nan if out of bounds)
        if np.any(x<=0) or np.any(x>=1) or np.any(self.alpha<=0) or np.any(self.beta<=0):
            return x*np.nan

        #Compute the gradient
        return (self.alpha - 1)/x + (self.beta-1)/(x-1)
        


import numpy as np
import cuqi
from cuqi.experimental.mcmc import ProposalBasedSampler
from cuqi.array import CUQIarray
from numbers import Number

class CWMH(ProposalBasedSampler):
    """Component-wise Metropolis Hastings sampler.

    Allows sampling of a target distribution by a component-wise random-walk
    sampling of a proposal distribution along with an accept/reject step.

    Parameters
    ----------

    target : `cuqi.distribution.Distribution` or lambda function
        The target distribution to sample. Custom logpdfs are supported by using
        a :class:`cuqi.distribution.UserDefinedDistribution`.
    
    proposal : `cuqi.distribution.Distribution` or callable method
        The proposal to sample from. If a callable method it should provide a
        single independent sample from proposal distribution. Defaults to a
        Gaussian proposal.  *Optional*.

    scale : float or ndarray
        Scale parameter used to define correlation between previous and proposed
        sample in random-walk.  *Optional*. If float, the same scale is used for
        all dimensions. If ndarray, a (possibly) different scale is used for
        each dimension.

    initial_point : ndarray
        Initial parameters. *Optional*

    callback : callable, optional
        A function that will be called after each sampling step. It can be useful for monitoring the sampler during sampling.
        The function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature is: `callback(sampler, sample_index, num_of_samples)`.

    kwargs : dict
        Additional keyword arguments to be passed to the base class 
        :class:`ProposalBasedSampler`.

    Example
    -------
    .. code-block:: python
        import numpy as np
        import cuqi
        # Parameters
        dim = 5 # Dimension of distribution
        mu = np.arange(dim) # Mean of Gaussian
        std = 1 # standard deviation of Gaussian

        # Logpdf function
        logpdf_func = lambda x: -1/(std**2)*np.sum((x-mu)**2)

        # Define distribution from logpdf as UserDefinedDistribution (sample
        # and gradients also supported as inputs to UserDefinedDistribution)
        target = cuqi.distribution.UserDefinedDistribution(
            dim=dim, logpdf_func=logpdf_func)

        # Set up sampler
        sampler = cuqi.experimental.mcmc.CWMH(target, scale=1)

        # Sample
        samples = sampler.sample(2000).get_samples()

    """

    _STATE_KEYS = ProposalBasedSampler._STATE_KEYS.union(['_scale_temp'])

    def __init__(self, target:cuqi.density.Density=None, proposal=None, scale=1,
                 initial_point=None, **kwargs):
        super().__init__(target, proposal=proposal, scale=scale,
                         initial_point=initial_point, **kwargs)
        
    def _initialize(self):
        if isinstance(self.scale, Number):
            self.scale = np.ones(self.dim)*self.scale
        self._acc = [np.ones((self.dim))] # Overwrite acc from ProposalBasedSampler with list of arrays

        # Handling of temporary scale parameter due to possible bug in old CWMH
        self._scale_temp = self.scale.copy()

    @property
    def scale(self):
        """ Get the scale parameter. """
        return self._scale

    @scale.setter
    def scale(self, value):
        """ Set the scale parameter. """
        if self._is_initialized and isinstance(value, Number):
            value = np.ones(self.dim)*value
        self._scale = value

    def validate_target(self):
        if not isinstance(self.target, cuqi.density.Density):
            raise ValueError(
                "Target should be an instance of "+\
                f"{cuqi.density.Density.__class__.__name__}")
        # Fail when there is no log density, which is currently assumed to be the case in case NaN is returned.
        if np.isnan(self.target.logd(self._get_default_initial_point(self.dim))):
            raise ValueError("Target does not have valid logd")
        
    def validate_proposal(self):
        if not isinstance(self.proposal, cuqi.distribution.Distribution):
            raise ValueError("Proposal must be a cuqi.distribution.Distribution object")
        if not self.proposal.is_symmetric:
            raise ValueError("Proposal must be symmetric")
        
    @property
    def proposal(self):
        if self._proposal is None:
            self._proposal = cuqi.distribution.Normal(
                mean=lambda location: location,
                std=lambda scale: scale,
                geometry=self.dim,
            )
        return self._proposal
    
    @proposal.setter
    def proposal(self, value):
        self._proposal = value

    def step(self):
        # Initialize x_t which is used to store the current CWMH sample
        x_t = self.current_point.copy()

        # Initialize x_star which is used to store the proposed sample by
        # updating the current sample component-by-component
        x_star = self.current_point.copy()

        # Propose a sample x_all_components from the proposal distribution
        # for all the components
        target_eval_t = self.current_target_logd
        if isinstance(self.proposal,cuqi.distribution.Distribution):
            x_all_components = self.proposal(
                location= self.current_point, scale=self.scale).sample()
        else:
            x_all_components = self.proposal(self.current_point, self.scale)

        # Initialize acceptance rate
        acc = np.zeros(self.dim)

        # Loop over all the components of the sample and accept/reject
        # each component update.
        for j in range(self.dim):
            # propose state x_star by updating the j-th component
            x_star[j] = x_all_components[j]

            # evaluate target
            target_eval_star = self.target.logd(x_star)

            # compute Metropolis acceptance ratio
            alpha = min(0, target_eval_star - target_eval_t)

            # accept/reject
            u_theta = np.log(np.random.rand())
            if (u_theta <= alpha) and \
               (not np.isnan(target_eval_star)) and \
               (not np.isinf(target_eval_star)):
                x_t[j] = x_all_components[j]
                target_eval_t = target_eval_star
                acc[j] = 1

            x_star = x_t.copy()

        self.current_target_logd = target_eval_t
        self.current_point = x_t

        return acc

    def tune(self, skip_len, update_count):
        # Store update_count in variable i for readability
        i = update_count

        # Optimal acceptance rate for CWMH
        star_acc = 0.21/self.dim + 0.23

        # Mean of acceptance rate over the last skip_len samples
        hat_acc = np.mean(self._acc[i*skip_len:(i+1)*skip_len], axis=0)

        # Compute new intermediate scaling parameter scale_temp
        # Factor zeta ensures that the variation of the scale update vanishes
        zeta = 1/np.sqrt(update_count+1)  
        scale_temp = np.exp(
            np.log(self._scale_temp) + zeta*(hat_acc-star_acc))

        # Update the scale parameter
        self.scale = np.minimum(scale_temp, np.ones(self.dim))
        self._scale_temp = scale_temp



""" This module defines messages (errors, warnings, etc) that are used throughout CUQIpy. The values of the following variables are not meant to be changed by the user."""

_disable_warning_msg = lambda module_name: "To disable "+\
"warnings for a given module or library, "+\
"you can use the method `warnings.filterwarnings`,"+\
" e.g.: warnings.filterwarnings(action='ignore', "+\
f"module=r'"+module_name+"')."
